import argparse
from ast import parse
import logging
import traceback
from typing import Callable, Any

from ..commands import initializer, environment, plan, deploy, run, cloud_output, local_development, destroy

parser = argparse.ArgumentParser(description="cdev cli")
subparsers = parser.add_subparsers(title="sub_command", description="valid subcommands")

def wrap_load_project(command: Callable) -> Callable[[Any], Any]:
    def wrapped_caller(*args, **kwargs):
        try:
            initializer.load_project(args)
        except Exception as e:
            print(f"Could not load the project to call {command}")
            print(e)
            print(traceback.format_exc())

            return

        command(args)

    return wrapped_caller


def wrap_load_and_initialize_project(command: Callable) -> Callable[[Any], Any]:
    def wrapped_caller(*args, **kwargs):
        try:
            initializer.load_and_initialize_project(args)
        except Exception as e:
            print(f"Could not load and initialize the project to call {command}")
            print(e)
            print(traceback.format_exc())
            return

        command(args)

    return wrapped_caller


CDEV_COMMANDS = [
    {
        "name": "plan",
        "help": "See the differences that have been made since the last deployment",
        "default": wrap_load_and_initialize_project(plan.plan_command_cli),
    },
    {
        "name": "init",
        "help": "Create a new project",
        "default": initializer.create_project_cli,
        "args": [
            {"dest": "name", "type": str, "help": "Name of the new project"},
            {"dest": "--template", "type": str, "help": "Name of the template for the new project"}
        ],
    },
    {
        "name": "develop",
        "help": "Open an interactive development environment",
        "default": wrap_load_and_initialize_project(local_development.develop_command_cli),
        "args": [
            {"dest": "--complex", "help": "run a simple follower instead of full development environment", "action":"store_true"}
        ]
    },
    {
        "name": "destroy",
        "help": "Destroy all the resources in the current environment",
        "default": wrap_load_and_initialize_project(destroy.destroy_command_cli)
    }, 

    {
        "name": "output",
        "help": "See the generated cloud output",
        "default": wrap_load_and_initialize_project(cloud_output.cloud_output_command_cli),
        "args": [
            {"dest": "cloud_output_id", "type": str, "help": "Id of the cloud output to display. ex: <component>.<ruuid>.<cdev_name>.<output_key>"}
        ]
    }, 
    {
        "name": "environment",
        "help": "Change and create environments for deployment",
        "default": wrap_load_and_initialize_project(environment.environment_cli),
        "subcommands": [
            {
                "command": "ls",
                "help": "Show all available environments",
                "args": [
                    {
                        "dest": "--all",
                        "help": "show more details",
                        "action": "store_true",
                    }
                ],
            },
            {
                "command": "set",
                "help": "Set the current working environment",
                "args": [
                    {
                        "dest": "env",
                        "type": str,
                        "help": "environment you want set as the new working environment",
                    }
                ],
            },
            {
                "command": "get",
                "help": "Get information about an environment",
                "args": [
                    {
                        "dest": "env",
                        "type": str,
                        "help": "environment you want info about",
                    }
                ],
            },
            {
                "command": "create",
                "help": "Create a new environment",
                "args": [
                    {
                        "dest": "env",
                        "type": str,
                        "help": "name of environment you want to create",
                    }
                ],
            },
            {
                "command": "settings_information",
                "help": "Get or Set information about the settings of an environment",
                "args": [
                    {
                        "dest": "--key",
                        "type": str,
                        "help": "The key to get or set.",
                    },
                    {
                        "dest": "--new-value",
                        "type": str,
                        "help": "New value of the variable. Must be used with --key.",
                    },
                    {
                        "dest": "--all",
                        "action": "store_true",
                        "help": "Set the value for all environments. Must be used with --new-value.",
                    }
                ],
            },
        ],
    },
    {
        "name": "deploy",
        "help": "Deploy a set of changes",
        "default": wrap_load_and_initialize_project(deploy.deploy_command_cli)
    },
    {
        "name": "run",
        "help": "This command is used to run user defined and resource functions.",
        "default": wrap_load_and_initialize_project(run.run_command_cli),
        "args": [
            {"dest": "subcommand", "help": "the user defined command to call"},
            {"dest": "args", "nargs": argparse.REMAINDER},
        ],
    },
]


def subcommand_function_wrapper(name, func):
    # This wraps a function so that is can be used for subcommands by basing the subcommand as the first arg to the function
    # then the remaining args as the second arg
    def inner(args):

        return func(name, args)

    return inner


def add_general_output_options(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--output",
        type=str,
        choices=["json", "plain-text", "rich"],
        help="change the type of output generated",
    )

    parser.add_argument(
        '-d', '--debug',
        help="Print debug log statements. This is mostly for development use",
        action="store_const", dest="loglevel", const=logging.DEBUG,
        default=logging.WARNING,
    )

    parser.add_argument(
        '-v', '--verbose',
        help="Print info log message. Use this to get a more detailed understanding of what is executing.",
        action="store_const", dest="loglevel", const=logging.INFO,
    )


for command in CDEV_COMMANDS:
    tmp = subparsers.add_parser(command.get("name"), help=command.get("help"))

    if command.get("subcommands"):
        tmp.set_defaults(func=subcommand_function_wrapper("", command.get("default")))
        t1 = tmp.add_subparsers()

        for subcommand in command.get("subcommands"):
            t2 = t1.add_parser(subcommand.get("command"), help=subcommand.get("help"))
            t2.set_defaults(
                func=subcommand_function_wrapper(
                    subcommand.get("command"), command.get("default")
                )
            )

            for arg in subcommand.get("args"):
                dest = arg.get("dest")
                arg.pop("dest")
                t2.add_argument(dest, **arg)

           
            add_general_output_options(t2)

    else:
        if command.get("args"):
            for arg in command.get("args"):
                dest = arg.get("dest")
                arg.pop("dest")
                tmp.add_argument(dest, **arg)

        tmp.set_defaults(func=command.get("default"))

        add_general_output_options(tmp)


args = parser.parse_args()
args.func(args)


from time import sleep
from typing import Callable, List
import boto3

AVAILABLE_SERVICES = set(["lambda", "s3", "dynamodb", "iam", "apigatewayv2", "sqs", "apigateway"])



def _get_boto_client(service_name, credentials=None, profile_name=None):

    # TODO readd this check after development is finished and we have the full list of services
    #if not service_name in AVAILABLE_SERVICES:
    #    return None

    if not credentials or not profile_name:
        return boto3.client(service_name)

    if credentials:
        return boto3.Session(
            aws_access_key_id=credentials.get("access_key"),
            aws_secret_access_key=credentials.get("secret_key")
        ).client(service_name)

    if profile_name:
        return boto3.Session(
            profile_name=profile_name
        ).client(service_name)


def get_boto_client(service_name):
    # TODO: Come back and make this settable from the workspace settings
    #if not cdev_settings.SETTINGS.get("CREDENTIALS"):
    return _get_boto_client(service_name)



def monitor_status(func: Callable, params: dict, previous_val, lookup_func: Callable) -> dict: 
    """
    This function is used to monitor the status of resources as they are created by aws. Alot of resources are created
    asynchronously by aws, which means the original create call returns before the actual resource is created. Therefor,
    we use this function to handle repeatedly calling a status function to check if the resource was created.   
    """

    MAX_RESOURCE_TIME = 600
    HEARTBEAT_PACE = 10

    loops = int(MAX_RESOURCE_TIME/HEARTBEAT_PACE)

    for _ in range(loops):
        rv = func(**params)

        if rv.get("ResponseMetadata").get("HTTPStatusCode") == 200:
            new_value = lookup_func(rv)
            
            if not new_value == previous_val:
                return rv
        
        sleep(HEARTBEAT_PACE)       


    return None


def run_client_function(service: str, function_name: str, args: dict, wait: dict = None):
    args_as_string = f"({service}, {function_name}, {args}, {wait})"
    
    rendered_client = _get_boto_client(service)
    method = getattr(rendered_client, function_name)

    if method:
        rv = method(**args)
        
    if wait:
        waiter = rendered_client.get_waiter(wait.get("name"))
        final_args = {
            "WaiterConfig":
                {
                    'Delay': 10,
                    'MaxAttempts': 60
                }
        }
        final_args.update( wait.get("args") )
        
        
        waiter.wait(**final_args)
        

    return rv


def aws_resource_wait(service: str, wait: dict):
    rendered_client = _get_boto_client(service)
    waiter = rendered_client.get_waiter(wait.get("name"))
    final_args = {
        "WaiterConfig":
            {
                'Delay': 10,
                'MaxAttempts': 60
            }
    }
    final_args.update( wait.get("args") )
    
    
    waiter.wait(**final_args)
    

def get_aws_region() -> str:
    return get_boto_client('s3').meta.region_name


def get_account_number() -> str:
    caller_info_rv =  run_client_function("sts", "get_caller_identity", {})
    return caller_info_rv.get('Account')


import boto3
import json
from typing import Any, Dict, List
from uuid import uuid4

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.default.resources.simple import relational_db 
from core.constructs.output_manager import OutputTask
from core.utils import hasher

from .. import aws_client



RUUID = "cdev::simple::relationaldb"


def _create_simple_relational_db(
    transaction_token: str, 
    namespace_token: str, 
    resource:  relational_db.simple_relational_db_model,
    output_task: OutputTask
) -> bool:

    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])
    cluster_name = f"cdev-relationaldb-{full_namespace_suffix}"
    
    output_task.update(comment=f'Creating DB {cluster_name}')

    rv = aws_client.run_client_function(
        "rds",
        "create_db_cluster",
        {
            "DatabaseName": resource.DatabaseName,
            "DBClusterIdentifier": cluster_name,
            "Engine": resource.Engine.value,
            "MasterUsername": resource.MasterUsername,
            "MasterUserPassword": resource.MasterUserPassword,
            "EnableHttpEndpoint": resource.EnableHttpEndpoint,
            "ScalingConfiguration": {
                "MinCapacity": resource.MinCapacity,
                "MaxCapacity": resource.MinCapacity,
                "AutoPause": resource.SecondsToPause == 0,
                "SecondsUntilAutoPause": resource.SecondsToPause
                if not resource.SecondsToPause == 0
                else 300,
            },
            "EngineMode": "serverless",  # Hard Code for this resource type
        },
    )

    output_task.update(comment=f'Creating Secrets for DB')    
    # Need to create secret for this information so that it can be used with the data api
    secret_rv = aws_client.run_client_function(
        "secretsmanager",
        "create_secret",
        {
            "Name": f"{cluster_name}-config",
            "SecretString": json.dumps(
                {
                    "dbInstanceIdentifier": cluster_name,
                    "engine": resource.Engine.value,
                    "host": rv.get("DBCluster").get("Endpoint"),
                    "port": rv.get("DBCluster").get("Port"),
                    "resourceId": rv.get("DBCluster").get("DbClusterResourceId"),
                    "username": resource.MasterUsername,
                    "password": resource.MasterUserPassword,
                }
            ),
        },
    )

    output_task.update(comment=f'Wating for DB to become available. This might take a minute.')
    aws_client.monitor_status(
        boto3.client("rds").describe_db_clusters,
        {
            "DBClusterIdentifier": cluster_name,
        },
        "creating",
        lambda x: x.get("DBClusters")[0].get("Status"),
    )

   
    output_info = {
        "endpoint": rv.get("DBCluster").get("Endpoint"),
        "secret_arn": secret_rv.get("ARN"),
        "cloud_id": rv.get("DBCluster").get("DBClusterArn"),
        "arn": rv.get("DBCluster").get("DBClusterArn"),
        "cdev_name": resource.name,
        "cluster_name": cluster_name
    }

    return output_info


def _update_simple_relational_db(
    transaction_token: str, 
    previous_resource: relational_db.simple_relational_db_model,
    new_resource: relational_db.simple_relational_db_model,
    previous_output: Dict,
    output_task: OutputTask
) -> bool:
    if not new_resource.Engine.value == previous_resource.Engine:
        raise Exception

    if not new_resource.MasterUsername == previous_resource.MasterUsername:
        raise Exception

    if not new_resource.DatabaseName == previous_resource.DatabaseName:
        raise Exception


    update_args = {"DBClusterIdentifier": previous_output.get('cluster_id')}

    scaling_config = {}

    update_configuration_secret = False
    if not new_resource.MasterUserPassword == previous_resource.MasterUserPassword:
        update_args["MasterUserPassword"] = new_resource.MasterUserPassword
        update_args["ApplyImmediately"] = True
        update_configuration_secret = True

    if not new_resource.EnableHttpEndpoint == previous_resource.EnableHttpEndpoint:
        update_args["EnableHttpEndpoint"] = new_resource.EnableHttpEndpoint

    if not new_resource.MaxCapacity == previous_resource.MaxCapacity:
        scaling_config["MaxCapacity"] = new_resource.MaxCapacity

    if not new_resource.MinCapacity == previous_resource.MinCapacity:
        scaling_config["MinCapacity"] = new_resource.MinCapacity

    if not new_resource.SecondsToPause == previous_resource.SecondsToPause:
        scaling_config["AutoPause"] = new_resource.SecondsToPause == 0
        scaling_config["SecondsUntilAutoPause"] = (
            new_resource.SecondsToPause if not new_resource.SecondsToPause == 0 else 300
        )

    if scaling_config:
        update_args["ScalingConfiguration"] = scaling_config

    output_task.update(comment=f'Updating DB Configuration')
    aws_client.run_client_function("rds", "modify_db_cluster", update_args)

    if update_configuration_secret:
        # Update the password in the secret
        secret_arn = previous_output.get("secret_arn")

        current_secret_rv = aws_client.run_client_function(
            "secretsmanager", "get_secret_value", {"SecretId": secret_arn}
        )

        current_secret_obj = json.loads(current_secret_rv.get("SecretString"))

        output_task.update(comment=f'Updating DB Secret Configuration')
        aws_client.run_client_function(
            "secretsmanager",
            "put_secret_value",
            {
                "SecretId": secret_arn,
                "SecretString": json.dumps(
                    {
                        "dbInstanceIdentifier": current_secret_obj.get(
                            "dbInstanceIdentifier"
                        ),
                        "engine": current_secret_obj.get("engine"),
                        "host": current_secret_obj.get("host"),
                        "port": current_secret_obj.get("port"),
                        "resourceId": current_secret_obj.get("resourceId"),
                        "username": current_secret_obj.get("username"),
                        "password": new_resource.MasterUserPassword,
                    }
                ),
            },
        )

    # None of the updatable values actually effect the output 
    return previous_output


def _remove_simple_relational_db(
    transaction_token: str, 
    previous_output: Dict,
    output_task: OutputTask
):
    cluster_id = previous_output.get('cluster_name')


    output_task.update(comment=f'[blink]Deleting DB. This could take a few minutes[/blink]')

    aws_client.run_client_function(
        "rds",
        "delete_db_cluster",
        {
            "DBClusterIdentifier": cluster_id,
            "SkipFinalSnapshot": True,
        },
    )

    # Update the password in the secret
    secret_arn = previous_output.get("secret_arn")

    output_task.update(comment=f'Deleting DB Secret')

    aws_client.run_client_function(
        "secretsmanager", "delete_secret", {"SecretId": secret_arn}
    )



def handle_simple_relational_db_deployment(
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_relational_db(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
        return _update_simple_relational_db(
            transaction_token,
            relational_db.simple_relational_db_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )

    elif resource_diff.action_type == Resource_Change_Type.DELETE:
        _remove_simple_relational_db(
            transaction_token,
            previous_output,
            output_task
        )

        return {}



#!/usr/bin/env python
import argparse
import logging
from typing import Callable, Any

from ..commands import (
    create_workspace,
    initialize_workspace,
    run,
    execute_frontend,
    create_resource_state,
    cloud_output
)
from ..constructs.workspace import Workspace

parser = argparse.ArgumentParser(description="cdev cli")
subparsers = parser.add_subparsers(title="sub_command", description="valid subcommands")


def wrap_initialize_workspace(command: Callable) -> Callable[[Any], Any]:
    def wrapped_caller(args):
        try:

            initialize_workspace.initialize_workspace_cli(args)
        except Exception as e:
            print(f"Could not initialize the workspace to call {command}")
            print(e)
            return

        command(args)

    return wrapped_caller


CDEV_COMMANDS = [
    {
        "name": "init",
        "help": "Create a new instance of a workspace",
        "default": create_workspace.create_workspace,
    },
    {
        "name": "plan",
        "help": "Create a new instance of a workspace",
        "default": wrap_initialize_workspace(execute_frontend.execute_frontend_cli),
        "args": [],
    },
    {
        "name": "create_resource_state",
        "help": "Create a new resource state",
        "default": wrap_initialize_workspace(
            create_resource_state.create_resource_state_cli
        ),
        "args": [],
    },
    {
        "name": "output",
        "help": "See the generated cloud output",
        "default": wrap_initialize_workspace(cloud_output.cloud_output_command_cli)
    }, 
    {
        "name": "run",
        "help": "This command is used to run user defined and resource functions.",
        "default": wrap_initialize_workspace(run.run_command),
        "args": [
            {"dest": "subcommand", "help": "the user defined command to call"},
            {"dest": "args", "nargs": argparse.REMAINDER},
        ],
    },
]


def subcommand_function_wrapper(name, subcommand):
    # This wraps a function so that is can be used for subcommands by basing the subcommand as the first arg to the function
    # then the remaining args as the second arg
    def inner(args):

        return command(subcommand, args)

    return inner


def add_general_output_options(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--output",
        type=str,
        choices=["json", "plain-text", "rich"],
        help="change the type of output generated",
    )

    parser.add_argument(
        '-d', '--debug',
        help="Print debug log statements. This is mostly for development use",
        action="store_const", dest="loglevel", const=logging.DEBUG,
        default=logging.WARNING,
    )

    parser.add_argument(
        '-v', '--verbose',
        help="Print info log message. Use this to get a more detailed understanding of what is executing.",
        action="store_const", dest="loglevel", const=logging.INFO,
    )


for command in CDEV_COMMANDS:
    tmp = subparsers.add_parser(command.get("name"), help=command.get("help"))

    if command.get("subcommands"):
        tmp.set_defaults(func=subcommand_function_wrapper("", command.get("default")))
        t1 = tmp.add_subparsers()

        for subcommand in command.get("subcommands"):
            t2 = t1.add_parser(subcommand.get("command"), help=subcommand.get("help"))
            t2.set_defaults(
                func=subcommand_function_wrapper(
                    subcommand.get("command"), command.get("default")
                )
            )

            for arg in subcommand.get("args"):
                dest = arg.get("dest")
                arg.pop("dest")
                t2.add_argument(dest, **arg)

            add_general_output_options(t2)

    else:
        if command.get("args"):
            for arg in command.get("args"):
                dest = arg.get("dest")
                arg.pop("dest")
                tmp.add_argument(dest, **arg)

        tmp.set_defaults(func=command.get("default"))

        add_general_output_options(tmp)


args = parser.parse_args()
args.func(args)



"""Structure that represents a distinct namespace of cloud resources


"""

from typing import Dict, List, Optional, Tuple
from .resource import Resource_Difference

from pydantic import BaseModel

from .components import ComponentModel


class Resource_State(BaseModel):
    """
    Parent class that describes a namespace that can store resource states via components and also be higher level states for other Resource States
    """

    name: str
    """
    Name for this resource state
    """

    uuid: str
    """
    Unique identifier for this state
    """

    components: Optional[List[ComponentModel]]
    """
    The list of components owned by this namespace
    """
    
    component_name_to_uuid: Optional[Dict[str, str]]
    """
    A dictionary from the uuid of a component to the component name
    """

    parent_uuid: Optional[str]
    """
    The parent namespace above this one
    """

    children: Optional[List[str]]
    """
    Child namespaces of this one
    """

    resource_changes: Optional[Dict[str, Tuple[str, Resource_Difference]]]
    """
    A dictionary of transaction tokens to a tuple of (component_name, diff)
    """

    failed_changes: Optional[Dict[str, Tuple[str, Resource_Difference, Dict]]]
    """
    A dictionary of transaction tokens to a tuple of (component_name, diff, error_info) for the changes that failed
    """


    def __init__(
        __pydantic_self__,
        name: str,
        uuid: str,
        components: List[ComponentModel] = [],
        component_name_to_uuid: Dict[str, str]={},
        parent_uuid: Optional[str] = None,
        children: List[str] = [],
        resource_changes: Dict[str, Tuple[str, Resource_Difference]] = {},
        failed_changes: Dict[str, Tuple[str, Resource_Difference, Dict]] = {},
        
    ) -> None:
        super().__init__(
            **{
                "name": name,
                "uuid": uuid,
                "components": components,
                "component_name_to_uuid": component_name_to_uuid,
                "parent_uuid": parent_uuid,
                "children": children,
                "resource_changes": resource_changes,
                "failed_changes": failed_changes
            }
        )



"""Set of constructs for making message queues

"""
from typing import Any

from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs, PermissionsAvailableMixin
from core.constructs.cloud_output import  Cloud_Output_Str, OutputType
from core.constructs.types import cdev_str_model

from core.default.resources.simple.events import Event, event_model
from core.default.resources.simple.iam import Permission

from core.utils import hasher

RUUID = "cdev::simple::queue"


######################
##### Events
######################
class queue_event_model(event_model):
    """Model representing a message queue event

    Arguments:
        batch_size: int
        batch_window: int
    """
    queue_arn: cdev_str_model   
    batch_size: int


class QueueEvent(Event):
    """Construct representing the Events from the Queue"""
    def __init__(self, queue_name: str, batch_size: int) -> None:

        if batch_size > 10000 or batch_size < 0:
            raise Exception

        self.cdev_queue_name = queue_name

        self.queue_arn = Cloud_Output_Str(
                name=queue_name, 
                ruuid=RUUID, 
                key='arn',
                type=OutputType.RESOURCE 
            )

        self.batch_size = batch_size


    def render(self) -> queue_event_model:
        return queue_event_model(
            originating_resource_name=self.cdev_queue_name,
            originating_resource_type=RUUID,
            hash=self.hash(),
            queue_arn=self.queue_arn.render(),
            batch_size=self.batch_size
        )


    def hash(self) -> str:
        return hasher.hash_list([
            self.cdev_queue_name,
            self.batch_size
        ])

######################
###### Permission
######################
class QueuePermissions:

    def __init__(self, resource_name: str) -> None:
        self.READ_QUEUE = Permission(
            actions=[
                "sqs:ReceiveMessage",
                "sqs:DeleteMessage",
                "sqs:GetQueueAttributes",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permission to poll for messages from the Queue"""

        self.WRITE_QUEUE = Permission(
            actions=[
                "sqs:SendMessage",
                "sqs:GetQueueAttributes",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permission to write messages to the Queue"""

        self.READ_AND_WRITE_QUEUE = Permission(
            actions=[
                "sqs:ReceiveMessage",
                "sqs:DeleteMessage",
                "sqs:GetQueueAttributes",
                "sqs:SendMessage",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permission to read and write messages to and from the Queue"""

        self.READ_EVENTS = Permission(
            actions=[
                "sqs:ReceiveMessage",
                "sqs:DeleteMessage",
                "sqs:GetQueueAttributes",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permission to receive events to and from the Queue"""

######################
##### Output
######################
class QueueOutput(ResourceOutputs):
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)

    @property
    def queue_name(self) -> Cloud_Output_Str:
        """The name of the generated queue"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='queue_name',
            type=self.OUTPUT_TYPE
        )

    @queue_name.setter
    def queue_name(self, value: Any):
        raise Exception


######################
##### Queue
######################
class queue_model(ResourceModel):
    """Model representing a Message Queue"""
    
    is_fifo: bool
    """Should the Queue guarantee ordering of messages"""


class Queue(PermissionsAvailableMixin, Resource):
    """A Message Queue"""

    @update_hash
    def __init__(self, cdev_name: str, is_fifo: bool = False, nonce: str = "") -> None:
        """
        args:
            cdev_name (str): Name of the resource.
            is_fifo (bool, default=False): If True, the Queue will guarantee ordering of messages.
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.
        """
        super().__init__(cdev_name, RUUID, nonce)

        self.is_fifo = is_fifo
        self.output = QueueOutput(cdev_name)
        self._event = None
        print(f"passing -> {cdev_name}")
        self.available_permissions: QueuePermissions = QueuePermissions(cdev_name)

    @property
    def is_fifo(self):
        """Should the Queue guarantee ordering of messages"""
        return self._is_fifo

    @is_fifo.setter
    @update_hash
    def is_fifo(self, value: bool):
        self._is_fifo = value

    def create_event_trigger(
        self, batch_size: int = 10
    ) -> QueueEvent:
        """Create an Event for the Queue that other resources can listen to

        Args:
            batch_size (int, optional): Size of message batch. Defaults to 10.

        Raises:
            Exception: _description_
            Exception: _description_

        Returns:
            QueueEvent
        """
        if self._event:
            raise Exception("Already created stream on this table. Use `get_stream()` to get the current stream.")
    
        # https://docs.aws.amazon.com/lambda/latest/dg/with-sqs.html#events-sqs-eventsource
        if self.is_fifo and batch_size > 10:
            raise Exception

        event = QueueEvent(
            queue_name=self.name,
            batch_size=batch_size,
        )

        self._event = event

        return event

    def get_event(self) -> QueueEvent:
        """Get the Event for this Queue

        Raises:
            Exception: _description_

        Returns:
            QueueEvent
        """
        if not self._event:
            raise Exception("Queue Event has not been created. Create a stream for this table using the `create_stream` function before calling this function.")

        return self._event

        
    def compute_hash(self):
        self._hash = hasher.hash_list([self.is_fifo, self.nonce])
        
    def render(self) -> queue_model:
        return queue_model(
            name=self.name,
            ruuid=self.ruuid,
            hash=self.hash,
            is_fifo=self.is_fifo,
        )



from core.constructs.workspace import Workspace, Workspace_State, initialize_workspace, load_workspace


from ..constructs.environment import Environment, environment_info
from core.default.cloudmapper import DefaultMapper

class local_environment(Environment):
    """
    A logically isolated instance of a project.
    """

    def __init__(self, info: environment_info) -> None:
        self.name = info.name
        self.workspace_info = info.workspace_info
        

    def get_workspace(self) -> Workspace:
        return Workspace.instance()

    def initialize_environment(self):
        ws = load_workspace(self.workspace_info)

        ws.set_state(Workspace_State.INITIALIZING)

        # Set the default mapper as a mapper in the workspace

        ws.add_mapper(DefaultMapper())

        initialize_workspace(ws, self.workspace_info.settings_info, self.workspace_info.config)

        ws.set_state(Workspace_State.INITIALIZED)

        




from argparse import ArgumentParser
import time, math, datetime

from typing import List, Dict
from boto3 import client

from core.constructs.commands import BaseCommand, OutputWrapper

from core.utils import hasher

from core.default.commands.function.utils import get_cloud_id_from_cdev_name




class show_logs(BaseCommand):

    help = """
        Get the logs of a deployed lambda function
    """

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "function_name", type=str, help="The function that you want to watch. Name must include component name. ex: comp1.demo_function"
        )
        parser.add_argument(
            "--watch",
            action="store_true",
            help="watch the logs. If this flag is passed, only --start_time flag is read",
        )
        parser.add_argument(
            "--tail", action="store_true", help="show the tail of the logs"
        )
        parser.add_argument(
            "--number",
            action="store_true",
            help="number of events to show. Must be used with --tail.",
        )

    def command(self, *args, **kwargs):

        full_function_name: str = kwargs.get("function_name")
        component_name = full_function_name.split('.')[0]
        function_name = full_function_name.split('.')[1]
        watch_val = kwargs.get("watch")
        tail_val = kwargs.get("tail")
        number_val = kwargs.get("number")

        cloud_name = get_cloud_id_from_cdev_name(component_name, function_name).split(":")[-1]

        if not cloud_name:
            self.stdout.write(f"Could not find function {function_name} in component {component_name}")
            return

        cloud_watch_group_name = f"/aws/lambda/{cloud_name}"

        if watch_val:
            _watch_log_group(cloud_watch_group_name, self.stdout)
            return

        cloud_watch_client = client("logs")
        print(cloud_watch_group_name)
        log_streams_rv = cloud_watch_client.describe_log_streams(
            logGroupName=cloud_watch_group_name,
            orderBy="LastEventTime",
        )

        streams = _get_needed_streams(log_streams_rv.get("logStreams"))

        for stream in streams:
            response = cloud_watch_client.get_log_events(
                logGroupName=cloud_watch_group_name,
                logStreamName=stream,
                startFromHead=True,
            )

            for event in response.get("events"):
                self.stdout.write(
                    f"{datetime.datetime.fromtimestamp(event.get('timestamp')/1000).strftime('%Y-%m-%d %H:%M:%S')} - {event.get('message')}"
                )


def _watch_log_group(group_name: str, stdout: OutputWrapper, args=None):
    cloud_watch_client = client("logs")
    _previous_refresh_time = int(
        (time.mktime(datetime.datetime.now().timetuple()) - 60) * 1000
    )
    events_hash = set()

    while True:
        try:

            log_streams_rv = cloud_watch_client.describe_log_streams(
                logGroupName=group_name,
                orderBy="LastEventTime",
            )

            streams = _get_needed_streams(
                log_streams_rv.get("logStreams"), _previous_refresh_time
            )

            for stream in streams:
                response = cloud_watch_client.get_log_events(
                    logGroupName=group_name,
                    logStreamName=stream,
                    startTime=_previous_refresh_time,
                    startFromHead=True,
                )

                for event in response.get("events"):
                    event_hash = hasher.hash_list(
                        [
                            event.get("timestamp"),
                            event.get("message"),
                            event.get("ingestionTime"),
                        ]
                    )

                    if not event_hash in events_hash:
                        stdout.write(
                            f"{datetime.datetime.fromtimestamp(event.get('timestamp')/1000).strftime('%Y-%m-%d %H:%M:%S')} - {event.get('message')}"
                        )

                    events_hash.add(event_hash)

                _previous_refresh_time = int(
                    (time.mktime(datetime.datetime.now().timetuple()) - 60) * 1000
                )

                time.sleep(0.2)
        except KeyboardInterrupt:
            return




def _get_needed_streams(streams: List[Dict], start_time: float = None) -> List[str]:
    stream_names = [x.get("logStreamName") for x in streams]

    return stream_names



import json
import os
from pydantic.types import DirectoryPath
import shutil
from typing import Dict, Tuple, List

from rich.prompt import Prompt
from cdev.default.project import local_project
from cdev.cli.logger import set_global_logger_from_cli

from core.constructs.backend import Backend, Backend_Configuration
from core.constructs.workspace import Workspace_Info
from core.default.backend import Local_Backend_Configuration, LocalBackend


from ..constructs.project import Project_State, check_if_project_exists, project_info


STATE_FOLDER = "state"
INTERMEDIATE_FOLDER = 'intermediate'
CDEV_FOLDER = ".cdev"
CDEV_PROJECT_FILE = "cdev_project.json"
CENTRAL_STATE_FILE = "central_state.json"
SETTINGS_FOLDER_NAME = 'settings'
DEFAULT_ENVIRONMENTS = ["prod", "stage", "dev"]
TEMPLATE_LOCATIONS = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'project_templates')


BASE_PROJECT_LOCATION = os.getcwd()

AVAILABLE_TEMPLATES = [
    'quick-start',
    'resources-test',
    'packages',
    'slack-bot'
]

def create_project_cli(args):
    config = args
    set_global_logger_from_cli(config.loglevel)

    if args.template:
        template_name = args.template

        if template_name not in AVAILABLE_TEMPLATES:
            print(f"{template_name} is not one of the available templates. {AVAILABLE_TEMPLATES}")
            return

    else:
        template_name = None

    create_project(args.name)
    print(f"Loading Template {template_name}")
    _load_template(template_name)
    

def _load_template(template_name: str):
    if not template_name:
        return

    template_folder_name = template_name.replace('-','_')


    if not template_folder_name in os.listdir(TEMPLATE_LOCATIONS):
        print(f"Could not finder template for {template_folder_name}")
        return

    template_location = os.path.join(TEMPLATE_LOCATIONS, template_folder_name)
    for x in os.listdir(template_location):
        
        full_location = os.path.join(template_location, x)
        if os.path.isdir(full_location):
            shutil.copytree(full_location, os.path.join(BASE_PROJECT_LOCATION, x))
        elif os.path.isfile(full_location):
            shutil.copyfile(full_location, os.path.join(BASE_PROJECT_LOCATION, x))


    print(f"Created Project From Template: {template_name}")

def create_project(project_name: str, base_directory: DirectoryPath = None):

    if not base_directory:
        base_directory = os.getcwd()

    if check_if_project_exists(base_directory):
        raise Exception("Project Already Created")


    _create_folder_structure(base_directory, DEFAULT_ENVIRONMENTS)

    base_settings_folder = os.path.join(base_directory, SETTINGS_FOLDER_NAME)
    


    backend_directory = os.path.join(base_directory, CDEV_FOLDER, STATE_FOLDER)
    backend_configuration = Local_Backend_Configuration(
        {
            "base_folder": backend_directory,
            "central_state_file": os.path.join(backend_directory, CENTRAL_STATE_FILE),
        }
    )

    new_project_info = project_info(
        project_name,
        environments=[],
        backend_info=backend_configuration,
        current_environment="",
    )

    project_info_location = os.path.join(base_directory, CDEV_FOLDER, CDEV_PROJECT_FILE)
    with open(project_info_location, "w") as fh:
        json.dump(new_project_info.dict(), fh, indent=4)

    new_project = local_project(project_info_location)

    new_project.initialize_project()

    # TODO restructure entire creating process so that this is more explicit
    base_dir = os.getcwd()

    for environment in DEFAULT_ENVIRONMENTS:

        environment_settings =  {
            "user_setting_module": [
                # set the settings modules as python modules
                os.path.relpath(os.path.join(base_settings_folder, f'base_settings.py'), start=base_dir)[:-3].replace('/',"."),
                os.path.relpath(os.path.join(base_settings_folder, f'{environment}_settings.py'), start=base_dir)[:-3].replace('/',".")
            ],
            "secret_dir":  os.path.relpath(os.path.join(base_settings_folder, f'{environment}_secrets'), start=base_dir),
        }

        new_project.create_environment(environment, environment_settings)
    

    new_project.set_state(Project_State.UNINITIALIZED)

    new_project.set_current_environment(DEFAULT_ENVIRONMENTS[-1])


def _create_folder_structure(base_directory: DirectoryPath, extra_settings: List[str]):
    """Create a skeleton file structure needed to make a project.

    Args:
        base_directory (DirectoryPath): [description]
        extra_settings (List[str]): [description]
    """

    if not os.path.isdir(os.path.join(base_directory, CDEV_FOLDER)):
        os.mkdir(os.path.join(base_directory, CDEV_FOLDER))

    if not os.path.isdir(os.path.join(base_directory, CDEV_FOLDER, STATE_FOLDER)):
        os.mkdir(os.path.join(base_directory, CDEV_FOLDER, STATE_FOLDER))

    if not os.path.isdir(os.path.join(base_directory, CDEV_FOLDER, INTERMEDIATE_FOLDER)):
        os.mkdir(os.path.join(base_directory, CDEV_FOLDER, INTERMEDIATE_FOLDER))


    base_settings_folder = os.path.join(base_directory, SETTINGS_FOLDER_NAME)

    if not os.path.isdir(base_settings_folder):
        os.mkdir(base_settings_folder)

    with open(os.path.join( base_settings_folder, f'base_settings.py'), 'w'):
        pass

    with open(os.path.join( base_settings_folder, f'__init__.py'), 'w'):
        pass
    
    for environment in extra_settings:
        with open( os.path.join( base_settings_folder, f'{environment}_settings.py'), 'w' ):
            pass

        os.mkdir(os.path.join(base_settings_folder, f'{environment}_secrets'))


def load_project(args):
    base_directory = os.getcwd()

    project_info_location = os.path.join(base_directory, CDEV_FOLDER, CDEV_PROJECT_FILE)

    local_project(project_info_location)


def load_and_initialize_project(args):
    base_directory = os.getcwd()

    project_info_location = os.path.join(base_directory, CDEV_FOLDER, CDEV_PROJECT_FILE)

    local_project(project_info_location).initialize_project()



from enum import Enum
import json
import os
from typing import Any, Dict, List, Optional, TypeVar, Tuple, Callable

from pydantic import BaseModel
from pydantic.types import DirectoryPath

from core.constructs.backend import Backend_Configuration
from core.constructs.mapper import CloudMapper
from core.constructs.components import Component
from core.constructs.types import F
from core.constructs.cloud_output import Cloud_Output
from core.constructs.settings import Settings, Settings_Info

from .environment import environment_info, Environment

_GLOBAL_PROJECT = None


class project_info(BaseModel):
    project_name: str
    environments: List[environment_info]
    backend_info: Backend_Configuration
    current_environment: Optional[str]

    def __init__(
        __pydantic_self__,
        project_name: str,
        environments: List[environment_info],
        backend_info: Backend_Configuration,
        current_environment: str = None,
       
    ) -> None:
        """
        Represents the data about a cdev project object:

        Parameters:
            project_name (str): Name of the project
            environments (List[environment_info]): The environments that are currently part of the project
            current_environment (str): The current environment
            
        """

        super().__init__(
            **{
                "project_name": project_name,
                "environments": environments,
                "backend_info": backend_info,
                "current_environment": current_environment,
            }
        )


class Project_State(str, Enum):
    UNINITIALIZED = "UNINITIALIZED"
    INITIALIZING = "INITIALIZING"
    INITIALIZED = "INITIALIZED"



def wrap_phases(phases: List[Project_State]) -> Callable[[F], F]:
    """
    Annotation that denotes when a function can be executed within the life cycle of a workspace. Throws excpetion if the workspace is not in the correct
    phase.
    """

    def inner_wrap(func: F) -> F:
        def wrapper_func(project: "Project", *func_posargs, **func_kwargs):

            current_state = project.get_state()
            if not current_state in phases:
                raise Exception(
                    f"Trying to call {func} while in project state {current_state} but need to be in {phases}"
                )

            else:

                return func(project, *func_posargs, **func_kwargs)

        return wrapper_func

    return inner_wrap


class Project:
    """
    Global class that defines the functionality for the Project object. The Project object is a global object that is created
    by Cdev to provide access to helpful API's for developers to use in their projects. The object is created and managed by
    the Cdev framework and follows a set of lifecycle rules (link to documentation).

    Developers can access the object in their projects by calling `Project.instance()`. When using the Project Object, developers
    should keep in mind the lifecycle rules around each API and how it fits into their project.

    """
    _settings: Settings

    @classmethod
    def instance(cls):
        """
        Method to retrieve the global instance of the Project object.
        """
        if not _GLOBAL_PROJECT:
            raise Exception("Currently No GLOBAL PROJECT OBJECT")

        return _GLOBAL_PROJECT

    @classmethod
    def set_global_instance(cls, project: "Project"):
        """
        Method to set the global Project object. Should only be used to as defined in the lifecycle of the Cdev process. (documentation)
        """
        global _GLOBAL_PROJECT
        _GLOBAL_PROJECT = project

    @classmethod
    def remove_global_instance(cls, caller: "Project"):
        """
        Method to reset the Global Project object. This should be the final cleanup step for a Cdev process.
        """
        global _GLOBAL_PROJECT

        if not _GLOBAL_PROJECT:
            raise Exception("Global Project is not set")

        if not _GLOBAL_PROJECT == caller:
            raise Exception("Only the current Project object can remove itself")

        _GLOBAL_PROJECT = None


    def set_name(self, name: str):
        """
        Set the name of this Project

        Args:
            name (str): name of the Project
        """
        raise NotImplementedError


    def get_name(self):
        """
        Get the name of this Project

        Returns:
            name (str)
        """
        raise NotImplementedError

    def get_state(self) -> Project_State:
        """
        Get the current lifecycle state of the Project.

        Returns:
            state (Project_State)
        """
        raise NotImplementedError

    def initialize_project(self):
        """
        Initialize the Project object so that it is ready to perform a given operation. Note that any configuration needed for the initialization
        process should be defined during the creation of the object, and set such that this function needs no input.
        """
        raise NotImplementedError

    def terminate_project(self):
        """
        Terminate the Project object as a cleanup operation after a given operation is complete. This should be the final step in the life cycle
        of an operation that need to initialize the project.
        """
        raise NotImplementedError

    def create_environment(self, environment_name: str, settings_files: List[str]):
        """
        Create a new environment for this project.
        """
        raise NotImplementedError

    def destroy_environment(self, environment_name: str):
        """
        Destroy an environment. This function should only be called when the project is in the Initialized state, so that
        any cloud resources in the environment will be destroyed.
        """
        raise NotImplementedError

    def get_all_environment_names(self) -> List[str]:
        """
        Get all the available environments in the Project.

        Returns:
            environment_names (List[str]): environments
        """
        raise NotImplementedError

    def get_current_environment_name(self) -> str:
        """
        Get the environment name that is currently active for this Project.

        Returns:
            environment_name (str)
        """
        raise NotImplementedError

    def set_current_environment(self, environment_name: str):
        """
        Change the currently active environment for this Project. This should only be called when the Project is
        in the Uninitialized state to prevent it from being called during operations that modify an environment.

        Arguments:
            environment_name (str): The environment name to switch to

        Raises:
            EnvironmentDoesNotExist
        """
        raise NotImplementedError

    def get_environment(self, environment_name: str) -> Environment:
        """
        Get the environment object for a specified environment name. Note that the environment will be in a state
        based on when this function is called within the Cdev lifecycle.

        Arguments:
            environment_name (str): The environment name to switch to

        Raises:
            EnvironmentDoesNotExist
        """
        raise NotImplementedError

    def get_current_environment(self) -> Environment:
        """
        Get the environment object for the currently active Environment. Note that the Environment will be in a state
        based on when this function is called within the Cdev lifecycle.

        Arguments:
            environment_name (str): The environment name to switch to

        Raises:
            EnvironmentDoesNotExist
        """
        raise NotImplementedError


    ############################
    ##### Settings
    ############################
    @property
    def settings(self) -> Settings:
        raise NotImplementedError

    
    @settings.setter
    def settings(self, value: Settings):
        raise NotImplementedError


    def get_settings_info(self, environment_name: str = None) -> Settings_Info:
        raise NotImplementedError

    
    def update_settings_info(self, new_value: Settings_Info, environment_name: str = None):
        raise NotImplementedError


    #######################
    ##### Display Output
    #######################

    def display_output(self, tag: str, output: Cloud_Output):
        """Display the output from a Resource or Reference after a process has completed

        Args:
            tag: A key value to display with the output
            output: The Cloud Output to render
        """
        raise NotImplementedError


    def render_outputs(self) -> List[Tuple[str, Any]]:
        """Render the output associated with the Workspace

        Returns:
            List[Tuple[str, Any]]: The List of outputs with their associated tag
        
        """
        raise NotImplementedError

    #################
    ##### Mappers
    #################
    def add_mapper(self, mapper: CloudMapper) -> None:
        """
        Add a CloudMapper to the project. The order that the Mappers are added to the Project defines the precedence give when
        determining which CloudMapper to use.

        Note that this function should only be called during the `Project Initialization` part of the Cdev lifecycle.

        Arguments:
            mapper (CloudMapper): The mapper to add
        """
        raise NotImplementedError

    def add_mappers(self, mappers: List[CloudMapper]) -> None:
        """
        Add a List of CloudMappers to the project. The order that the Mappers are added to the Project defines the precedence
        give when determining which CloudMapper to use.

        Note that this function should only be called during the `Project Initialization` part of the Cdev lifecycle.

        Arguments:
            mappers (List[CloudMapper]): The mapper to add
        """
        raise NotImplementedError

    def get_mappers(self) -> List[CloudMapper]:
        """
        Return the List of CloudMappers for this Project.

        Note that this function should only be called during the `Project Initialized` part of the Cdev lifecycle.

        Returns:
            mappers (List[CloudMapper]): mappers for this Project
        """
        raise NotImplementedError

    def get_mapper_namespace(self) -> Dict[str, CloudMapper]:
        """
        Return the Dictionary that maps Resource ID's (ruuid) to the mapper that will be used to deploy the resource into the cloud.

        Note that this function should only be called during the `Project Initialized` part of the Cdev lifecycle.

        Returns:
            ruuid_to_mapper (Dict[str, CloudMapper]): Resource ID to CloudMapper
        """
        raise NotImplementedError

    #################
    ##### Commands
    #################
    def add_command(self, command_location: str):
        """
        Add a Command Location to the Project. The order that the Command is added to the Project defines the precedence
        give when searching for Commands. Command Locations should adhere to the defined form to ensure that they can be
        found within the Project.

        Note that this function should only be called during the `Project Initialization` part of the Cdev lifecycle.

        Arguments:
            command_location (Command): The command location to add
        """
        raise NotImplementedError

    def add_commands(self, command_locations: List[str]):
        """
        Add a List of Command Locations to the Project. The order that the Commands are added to the Project defines the precedence
        give when searching for a Command. Command Locations should adhere to the defined form to ensure that they can be
        found within the Project.

        Note that this function should only be called during the `Project Initialization` part of the Cdev lifecycle.

        Arguments:
            command_locations (Command): The command location to add
        """
        raise NotImplementedError

    def get_commands(self) -> List[str]:
        """
        Get the Command Locations for this Project.

        Note that this function should only be called during the `Project Initialized` part of the Cdev lifecycle.

        Returns:
            command_locations (List[str])
        """
        raise NotImplementedError

    #################
    ##### Components
    #################
    def add_component(self, component: Component):
        """
        Add a Component to the Project. Components are used to determine the desired state of the Project. They should represent
        a logical separation for the Resources in a project.

        Note that this function should only be called during the `Project Initialization` part of the Cdev lifecycle.

        Arguments:
            component (Component): Component to add
        """
        raise NotImplementedError

    def add_components(self, components: List[Component]):
        """
        Add a List of Components to the Project. Components are used to determine the desired state of the Project. They
        should represent a logical separation for the Resources in a project.

        Note that this function should only be called during the `Project Initialization` part of the Cdev lifecycle.

        Arguments:
            component (Component): Component to add
        """
        raise NotImplementedError

    def get_components(self) -> List[Component]:
        """
        Return the Components for this Project.

        Note that this function should only be called during the `Project Initialized` part of the Cdev lifecycle.

        Returns:
            components (List[Component])
        """
        raise NotImplementedError


def check_if_project_exists(base_directory: DirectoryPath) -> bool:
    """
    This function checks for an existing Cdev project at the given directory. A Cdev project is defined by the existence of a valid
    'cdev_project.json' file (can be loaded as project_info) at the location of <base_directory>/.cdev/.

    Arguments:
        base_directory: Directory to start search from
    """

    CDEV_FOLDER = ".cdev"
    CDEV_PROJECT_FILE = "cdev_project.json"

    if not os.path.isdir(base_directory):
        raise Exception(f"Given base directory is not a directory {base_directory}")

    if not os.path.isdir(os.path.join(base_directory, CDEV_FOLDER)):
        return False

    if not os.path.isfile(os.path.join(base_directory, CDEV_FOLDER, CDEV_PROJECT_FILE)):
        return False

    try:
        with open(
            os.path.join(base_directory, CDEV_FOLDER, CDEV_PROJECT_FILE), "r"
        ) as fh:
            project_info(**json.load(fh))

        return True

    except Exception as e:
        return False



import ast
import os
from sys import modules, version_info

from typing import Dict, List, Set, Tuple, Union
from pydantic.types import FilePath


from .parser_objects import *
from .parser_exceptions import InvalidParamError, CouldNotParseFileError, CdevFileNotFoundError, InvalidDataError

EXCLUDED_SYMBOLS = set(["print"])

line_type = Union[ast.stmt, ast.expr]

def parse_line_numbers(objs: List[line_type], final_line: int) -> List[Tuple[ line_type, Tuple[int,int] ] ]:
    rv = []
    previous_info = None
    for node in objs:
        #_tmp_global_information[node] = [node.lineno, node.end_lineno]
        if previous_info:
            previous_name, previous_start_line = previous_info
            line_nos = (previous_start_line, node.lineno-1)
            rv.append( (previous_name , line_nos) )

        previous_info = (node, node.lineno)

    # Set the last line of the last global statement to the last line of the file
    previous_name, previous_start_line = previous_info
    line_nos = (previous_start_line, final_line)
    rv.append( (previous_name , line_nos) )

    return rv


def find_ending_line(file_info_obj: file_information, start_line: int, end_line: int, line_start_match: str, ):
    for i in range(start_line+1, end_line):
        tmp = file_info_obj.get_lines_of_source_code(i, i)
        
        if tmp.startswith(line_start_match):
            return i

    raise Exception("Should not end loop without returning")


def _get_global_variables_in_symboltable(table):
    if not isinstance(table, symtable.SymbolTable):
        raise InvalidParamError(f"Invalid Param in _get_functions_in_symboltable: Expected symtable.SymbolTable got {type(table)}")

    symbols = table.get_symbols()

    rv_variables = set()

    for sym in symbols:
        # Check all symbols to see if they are not a namespace. Check if the symbol is a global.
        if not sym.is_namespace():
            if sym.is_global():
                rv_variables.add(sym.get_name())

    return rv_variables


def _get_imported_variables_in_symboltable(table):
    # TODO Change error that is raised
    if not isinstance(table, symtable.SymbolTable):
        raise InvalidParamError(f"Invalid Param in _get_functions_in_symboltable: Expected symtable.SymbolTable got {type(table)}")

    symbols = table.get_symbols()

    rv_variables = set()

    for sym in symbols:
        # Check all symbols to see if they are not a namespace. If not then they are a normal symbols for this symboltable.
        if not sym.is_namespace():
            if sym.is_imported():
                rv_variables.add(sym.get_name())

    return rv_variables


def _get_local_variables_in_symboltable(table):
    if not isinstance(table, symtable.SymbolTable):
        raise InvalidParamError(f"Invalid Param in _get_functions_in_symboltable: Expected symtable.SymbolTable got {type(table)}")

    symbols = table.get_symbols()

    rv_variables = set()

    for sym in symbols:
        # Check all symbols to see if they are not a namespace. If not then they are a normal gloabl variable for this symboltable.
        if not sym.is_namespace():
            if not sym.is_imported() and not sym.is_free(
            ) and not sym.is_global():
                rv_variables.add(sym.get_name())

    return rv_variables


def _get_functions_in_symboltable(table):
    if not isinstance(table, symtable.SymbolTable):
        raise InvalidParamError(f"Invalid Param in _get_functions_in_symboltable: Expected symtable.SymbolTable got {type(table)}")

    symbols = table.get_symbols()

    rv_functions = set()

    for sym in symbols:
        # Check all symbols to see if they are a namespace (only functions and classes are namespaces)
        if sym.is_namespace():
            for ns in sym.get_namespaces():
                # A symbol can have multiple bindings so we need to see if any of them are functions
                if isinstance(ns, symtable.Function):
                    rv_functions.add(sym.get_name())
                    continue

    return rv_functions


def _get_global_class_definitions_in_symboltable(table):
    if not isinstance(table, symtable.SymbolTable):
        raise InvalidParamError(f"Invalid Param in _get_functions_in_symboltable: Expected symtable.SymbolTable got {type(table)}")

    symbols = table.get_symbols()

    rv_class_defs = set()

    for sym in symbols:
        # Check all symbols to see if they are a namespace (only functions and classes are namespaces)
        if sym.is_namespace():
            for ns in sym.get_namespaces():
                # A symbol can have multiple bindings so we need to see if any of them are functions
                if isinstance(ns, symtable.Class):
                    rv_class_defs.add(sym.get_name())
                    continue

    return rv_class_defs


def _generate_global_statement(file_info_obj: file_information, ast_node: ast.NodeVisitor, line_info: List[str], remove_annotation_functions: Set[str] = set()):
    # This function is used to determine the type of global statement the node is and create the corresponding global statement obj

    last_line = line_info[1]

    manual_include = False
    manual_include_sym = ""

    # Need to adjust the starting line if using python3.8 or greater
    start_line = line_info[0]

    if (isinstance(ast_node, ast.FunctionDef) and ast_node.decorator_list):
        # Since python < 3.8 does not include the end_line info on an ast.node, we need to handle removing the top annotation differently 
        if version_info < (3,8):
            if not ast_node.name in remove_annotation_functions:
                start_line = ast_node.decorator_list[0].lineno

            else:
                # Since we have no way of know when the first decorator ends because there is not end_lineno property,
                # we have to check each line below it to see if it is either the function declaration ('def ') or
                # the next annotation definition ('@')
                match_line = "def " if len(ast_node.decorator_list) == 1 else "@"

                # search for the next definition based on the match_line and search at longest til you get to the first line of the body of the function
                start_line = find_ending_line(file_info_obj, ast_node.decorator_list[0].lineno, ast_node.body[0].lineno, match_line)
        else:
            # for python>=3.8
            start_line = ast_node.decorator_list[0].end_lineno + 1 if ast_node.name in remove_annotation_functions else ast_node.decorator_list[0].lineno

    # Look one line above the start for a manual include
    for k in file_info_obj.include_overrides_lineno:
        if file_info_obj.include_overrides_lineno.get(k) == start_line-1:
            manual_include = True
            manual_include_sym = k
            break

    tmp_src_code = file_info_obj.get_lines_of_source_code(
        start_line, last_line)

    tmp_symbol_table = symtable.symtable(tmp_src_code,
                                         file_info_obj.file_location, 'exec')


    need_to_check_function = False
    need_to_check_class_def = False

    for sym in tmp_symbol_table.get_symbols():
        if sym.is_namespace():
            if isinstance(sym.get_namespaces()[0], symtable.Function):
                need_to_check_function = True
            elif isinstance(sym.get_namespaces()[0], symtable.Class):
                need_to_check_class_def = True
        

    if need_to_check_function:
        single_symbol = tmp_symbol_table.get_symbols()[0]
        
        if single_symbol.is_namespace():
            if isinstance(single_symbol.get_namespaces()[0],
                          symtable.Function):
                    
                name = single_symbol.get_name()
                
                function_symbols = set([x.get_name() for x in single_symbol.get_namespaces()[0].get_symbols()])

                if isinstance(ast_node, ast.FunctionDef):
                    if ast_node.decorator_list:
                        # If this is a function with annotations then we should include all the symbols from the top
                        # level tmp symbol table so that the variables used in the annotations are captured.
                        function_symbols.update(set([x.get_name() for x in tmp_symbol_table.get_symbols()]))


                fs = FunctionStatement(ast_node, [start_line, last_line],
                                       list(function_symbols), name)
                file_info_obj.add_global_function(name, fs)

                if manual_include:
                    file_info_obj.include_overrides_glob[manual_include_sym] = fs

                return

    if need_to_check_class_def:

        class_def_table = tmp_symbol_table.get_children()[0]

        class_def = ClassDefinitionStatement(ast_node, [start_line, last_line], [x.get_name() for x in class_def_table.get_symbols()] , class_def_table.get_name())
        file_info_obj.add_class_definition(class_def_table.get_name(), class_def)
        return 

    ts = {s.get_name() for s in tmp_symbol_table.get_symbols()}

    used_imported_symbols = file_info_obj.imported_symbols.intersection(ts)
    if len(used_imported_symbols) > 0:
        # This statement uses an imported symbol so we need to check it to see if it is the import statement
        for n in ast.walk(ast_node):
            #print(n)
            if isinstance(n, ast.Import):
                #print(f"IMPORT: {n}")
                #print(f"FIELDS { [(cn.name, cn.asname)  for cn in n.names] }")
                for imprt in n.names:
                    #print(f"FIELDS: {imprt.name}; {imprt.asname}")
                    if not imprt.asname:
                        asname = imprt.name
                    else:
                        asname = imprt.asname

                    imp_statement = ImportStatement(ast_node,
                                                    [start_line, last_line],
                                                    [x.get_name() for x in tmp_symbol_table.get_symbols()], asname,
                                                    imprt.name)

                    #print(f"{ast.dump(n)} -> {asname}?{imprt.name}")
                    file_info_obj.add_global_import(imp_statement)
                    continue

            if isinstance(n, ast.ImportFrom):
                for imprt in n.names:
                    if not imprt.asname:
                        asname = imprt.name
                    else:
                        asname = imprt.asname
                    

                    if n.level > 0:
                        if not n.module:
                            asmodule = f"{'.' * n.level}{asname}"
                        else:
                            asmodule = f"{'.' * n.level}{n.module}"
                            

                    else:
                        asmodule =  n.module

                    #print(f"{ast.dump(n)} -> {asname}?{asmodule}")
                    imp_statement = ImportStatement(ast_node,
                                                    [start_line, last_line],
                                                    [x.get_name() for x in tmp_symbol_table.get_symbols()], asname,
                                                    asmodule)

                    file_info_obj.add_global_import(imp_statement)
                continue

    global_statement_obj = GlobalStatement(ast_node, [start_line, last_line],
                                           [x.get_name() for x in tmp_symbol_table.get_symbols()])


    
    file_info_obj.add_global_statement(global_statement_obj)

    if manual_include:
        file_info_obj.include_overrides_glob[manual_include_sym] = global_statement_obj



def get_file_information(file_path: str, include_functions: List[str] =[], function_manual_includes: Dict[str, str]={}, global_manual_includes: List[str]=[], remove_top_annotation: bool =False):
    if not os.path.isfile(file_path):
        raise CouldNotParseFileError(CdevFileNotFoundError(
            f"cdev_parser: could not find file at -> {file_path}"))


    file_info_obj = file_information(file_path)

    symbol_table = file_info_obj.get_symbol_table()

    # Need to get some basic information about the global namespace in the file. This includes:
    # - all defined functions (_get_functions_in_symboltable)
    # - variables defined in the global namespace (_get_local_variables_in_symboltable)
    # - imported symbols (_get_imported_variables_in_symboltable)
    # - global variables (_get_global_variables_in_symboltable)

    try:
        functions = _get_functions_in_symboltable(symbol_table)

        file_info_obj.function_names = functions

        local_variables = _get_local_variables_in_symboltable(symbol_table)

        imported_symbols = _get_imported_variables_in_symboltable(symbol_table)

        global_symbols = _get_global_variables_in_symboltable(symbol_table)

        class_definitions = _get_global_class_definitions_in_symboltable(symbol_table)
    except InvalidParamError as e:
        raise CouldNotParseFileError(e)
        return

    for item in functions.union(local_variables).union(imported_symbols).union(
            global_symbols).union(class_definitions):
        file_info_obj.symbol_to_statement[item] = set()

    file_info_obj.set_imported_symbols(imported_symbols)

    # Need to get the line range of each global statement, but this requires looping over all the global nodes in the ast
    # in the top level of the file. 
    # 
    # Python3.8 introduced node.end_lineno to ast nodes, which makes it easier to get the ending line info
    # 
    # To support earlier versions of python (<3.8) we must use the starting line of
    # the next statement as the last line of previous node. We are going to store this info in a tmp dict then
    # once we have all the information create the objects

    file_ast = file_info_obj.get_ast()

    # dict<ast.node, [line1,line2]>
    _tmp_global_information = {}

    # sys.version_info
    if version_info < (3,8):
        _tmp_global_information = dict(parse_line_numbers(file_ast.body, file_info_obj.get_file_length() + 1 ))
    else:
        for node in file_ast.body:
            _tmp_global_information[node] = [node.lineno, node.end_lineno]



    remove_function_annotations = set(include_functions) if remove_top_annotation else set()
    # Now that the information has been collected for the global statements, we can create the actual objs and add
    # them to the file_info_obj
    for node, line_info in _tmp_global_information.items():
        _generate_global_statement(file_info_obj, node, line_info, remove_function_annotations)


    # Build a two-way binding of a symbol to statement and a statement to a symbol.
    # This information is needed to get all dependencies of symbols, which is needed
    # to reconstruct just the necessary lines
    for i in file_info_obj.get_global_statements():
        syms = i.get_symbols()
        
        for sym in syms:
            if not sym in file_info_obj.symbol_to_statement:
                continue

            # Some symbols do no create a dependency so they should not be added
            if sym in EXCLUDED_SYMBOLS:
                continue

            # Add all the symbols in the statement to the list for this global statement. This means when this global object
            # is used we need to include all of these symbols
            if i in file_info_obj.statement_to_symbol:
                tmp = file_info_obj.statement_to_symbol.get(i)
                tmp.add(sym)
                file_info_obj.statement_to_symbol[i] = tmp

            # Since this global statement is a top level function... the code that effects this symbol within it will only execute if the
            # function itself is called. Therefor, other functions can use this symbol without depending on this function.
            if i.get_type() == GlobalStatementType.FUNCTION:
                continue

            tmp = file_info_obj.symbol_to_statement.get(sym)
            tmp.add(i)
            file_info_obj.symbol_to_statement[sym] = tmp


        # if this global object is a function then we need to add the global object to its own symbol name dependency list. That way
        # if another global statement calls this function we include this global statement (which is the definition of the function) 
        if i.get_type() == GlobalStatementType.FUNCTION:
            tmp = file_info_obj.symbol_to_statement.get(i.get_function_name())
            tmp.add(i)
            file_info_obj.symbol_to_statement[i.get_function_name()] = tmp

        if i.get_type() == GlobalStatementType.CLASS_DEF:
            tmp = file_info_obj.symbol_to_statement.get(i.get_class_name())
            tmp.add(i)
            file_info_obj.symbol_to_statement[i.get_class_name()] = tmp


    # If a list of functions was not included then parse all top level functions
    if not include_functions:
        include_functions = file_info_obj.global_functions.keys()


    # manual includes is a dictionary from function name to global statement
    for function_name in include_functions:
        # Create new parsed function obj
        p_function = parsed_function(function_name)

        # Start with a set that already includes the symbol for itself
        already_included_symbols = set([function_name])

        # All functions will need to start by including the actual function body
        # Some functions will also need to include the manually added statements
        needed_global_objects = set([file_info_obj.global_functions.get(function_name)])

        # if the function has any manual statement includes
        # function_manual_includes is the dict from function name to the name of the manual include
        if function_name in function_manual_includes:
            for include in function_manual_includes.get(function_name):
                if not include in file_info_obj.include_overrides_glob:
                    raise InvalidDataError(f"""function_manual_includes param data issue: Trying to manually include '{include}' for function '{function_name} but '{include}' is not present in file""")
                    

                needed_global_objects.add(file_info_obj.include_overrides_glob.get(include))
        
        # global_manual_includes is a list of the names of manual includes that need to be added to every function
        for include in global_manual_includes:
            if not include in file_info_obj.include_overrides_glob:
                raise InvalidDataError(f"""function_manual_includes param data issue: Trying to manually include '{include}' as a global manual include but '{include}' is not present in file""")

            needed_global_objects.add(file_info_obj.include_overrides_glob.get(include))

        
        next_symbols = set()
        already_included_global_obj = set()

        
        for global_object in needed_global_objects:
            # Add the functions lines to the parsed function
            p_function.add_line_numbers(global_object.get_line_no())

            # Start with a set that already includes the global statement for itself
            already_included_global_obj.add(global_object)

            # next symbols is the set of symbols that need their dependant global statements
            next_symbols = next_symbols.union(file_info_obj.statement_to_symbol.get(global_object))

            # Some symbols won't be included in the mapping because they are excluded, but they need to be kept track of to look at imports
            all_used_symbols = set(global_object.get_symbols())

        keep_looping = True
        remaining_symbols = set()

        while keep_looping:
            # If there are no more symbols than break the loop
            


            # Add the previous iterations symbols as already seen so they are not readded
            already_included_symbols = already_included_symbols.union(
                remaining_symbols)
            remaining_symbols = next_symbols

            if not next_symbols:
                break

            next_symbols = set()

            for sym in remaining_symbols:
                if sym in already_included_symbols:
                    # If we have already added this symbol than we have all of its dependencies
                    continue

                # Add all dependant global statements to this symbol needs to the need lines for the function
                for glob_obj in file_info_obj.symbol_to_statement.get(sym):
                    if glob_obj in already_included_global_obj:
                        #IF this global statement has already been added continue
                        continue

                    # Add the lines numbers and place it in the already included set
                    p_function.add_line_numbers(glob_obj.get_line_no())
                    already_included_global_obj.add(glob_obj)

                    # include this statements
                    all_used_symbols = all_used_symbols.union(set(glob_obj.get_symbols()) )

                    # Get the set of symbols that this statement depends on
                    set_symbols = file_info_obj.statement_to_symbol.get(
                        glob_obj)
                    # Remove the symbols that have already been included
                    new_needed_symbols = set_symbols.difference(
                        already_included_symbols)
                    actual_new_needed_symbols = new_needed_symbols.difference(
                        remaining_symbols)

                    # Add the actually needed new symbols to the set of symbols to be looped over next
                    next_symbols = next_symbols.union(
                        actual_new_needed_symbols)

        #print(f"all pkg in file -> {file_info_obj.imported_symbol_to_global_statement}")
        for symbol in all_used_symbols:
            if symbol in file_info_obj.imported_symbol_to_global_statement and not symbol in EXCLUDED_SYMBOLS:
                p_function.add_import(
                    file_info_obj.imported_symbol_to_global_statement.get(symbol)
                )

        
        
        #finally add the parsed function object to the file info
        file_info_obj.add_parsed_functions(p_function)

    return file_info_obj


_individual_file_cache = {}
def _get_individual_files_imported_symbols(file_location):
    if file_location in _individual_file_cache:
        return _individual_file_cache.get(file_location)

    rv = set()
    
    with open(file_location, 'r') as fh:
        src_code = fh.readlines()
        fh.seek(0)

    ast_rep = ast.parse("".join(src_code))

    

    for node in ast.walk(ast_rep):
        
        if isinstance(node, ast.Import):
            for pkg_name in node.names:
                if not pkg_name.asname:
                    asname = pkg_name.name
                else:
                    asname = pkg_name.asname
                rv.add((asname, None))
    
        elif isinstance(node, ast.ImportFrom):
            # if the import has levels > 0 and no module it is a local referenced package and will already be included
            # if not then it is a package on the PYTHONPATH and we need to add it as a dependency
            
            
            for pkg_name in node.names:
                
                if node.level > 0:
                    if not node.module:
                        asmodule = f"{'.' * node.level}{pkg_name.name}"
                    else:
                        asmodule = f"{'.' * node.level}{node.module}"

                    rv.add((asmodule, file_location))

                else:
                    asmodule =  node.module
                    rv.add((asmodule, None))

    _individual_file_cache[file_location] = rv
                
    return rv




def get_file_imported_symbols(file_loc: FilePath):
    # Walk the whole dir/children importing all python files and then searching their symbol tree to find import statements
    rv = _get_individual_files_imported_symbols(file_loc)

    return rv





"""Utilities for creating a new resource state

"""
import os

from core.constructs import workspace as cdev_workspace
from core.constructs.backend import Backend_Configuration
from core.constructs.settings import Settings_Info

from core.default.workspace import local_workspace_manager


def create_workspace(args):
    base_project_dir = os.getcwd()
    manager = local_workspace_manager(base_project_dir)

    if manager.check_if_workspace_exists():
        print("Workspace already initialized")

    workspace_info = cdev_workspace.Workspace_Info(
        "core.default.workspace",
        "local_workspace",
        Settings_Info(
            base_class="core.constructs.settings.Settings"
        ),
        {
            "backend_configuration": {
                "python_module": "core.default.backend",
                "python_class": "LocalBackend",
                "config": {
                    "base_folder": base_project_dir,
                    "central_state_file": "centralstate.json",
                },
            },
            "initialization_file": "cdev_project",

        },
    )

    manager.create_new_workspace(workspace_info)

    return



from cdev.constructs.project import Project
from cdev.default.output_manager import CdevOutputManager
from cdev.cli.logger import set_global_logger_from_cli

from core.commands.deploy_differences import execute_deployment



def deploy_command_cli(args):
    config = args[0]
    set_global_logger_from_cli(config.loglevel)
    deploy_command(args)


def deploy_command(args):

    output_manager = CdevOutputManager()
    myProject = Project.instance()

    ws = myProject.get_current_environment().get_workspace()

    execute_deployment(ws, output_manager)




"""Command for uploading objects to a Bucket

"""
from argparse import ArgumentParser
from imp import source_from_cache
import boto3
import os
import mimetypes

from core.constructs.commands import BaseCommand, OutputWrapper
from core.default.resources.simple.object_store import Bucket, bucket_model


from . import utils


class cp(BaseCommand):
    help = """
    Command to mirror s3 `cp` command.
    
    """

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "source",
            type=str,
            help="The source of the file. Must either be a local path or bucket location (bucket://<component>.<name>/<path>)",
        )
        parser.add_argument(
            "destination",
            type=str,
            help="The destination of the file. Must either be a local path or bucket location (bucket://<component>.<name>/<path>)",
        )

        parser.add_argument(
            "--recursive",
            action="store_true",
            help="The destination of the file. Must either be a local path or bucket location (bucket://<component>.<name>/<path>)",
        )

    def command(self, *args, **kwargs):
        
        source_raw = kwargs.get("source")
        destination_raw = kwargs.get("destination")

        is_recursive = kwargs.get("recursive")


        if utils.is_valid_remote(source_raw):
            is_source_remote = True
            source = utils.parse_remote_location(source_raw)
            
        elif os.path.isfile(source_raw):
            is_source_remote = False
            source = source_raw

        elif os.path.isdir(source_raw) and is_recursive:
            is_source_remote = False
            source = source_raw

        else:
            raise Exception(f"Source {source_raw} is neither a valid location on the file system or a valid remote location")



        if utils.is_valid_remote(destination_raw):
            is_destination_remote = True
            destination = utils.parse_remote_location(destination_raw)
        
        elif os.path.isdir(destination_raw):
            is_destination_remote = False
            destination = destination_raw

        elif os.path.isdir(os.path.dirname(destination_raw)):
            is_destination_remote = False
            destination = destination_raw

        

        else:
            raise Exception(f"Destination {destination_raw} is neither a valid location on the file system or a valid remote location")


        if not is_destination_remote and not is_source_remote:
            raise Exception(f"Both destination ({destination}) and source ({source}) are file system locations")


        s3 = boto3.resource("s3")
        

        if is_destination_remote and not is_source_remote:
            # Remote destination and local file system. 
            # Copying up
            cloud_output = utils.get_cloud_output_from_cdev_name(destination.component_name, destination.cdev_bucket_name)
            bucket_name = cloud_output.get('bucket_name')
            

            bucket = s3.Bucket(bucket_name)

            if destination.path:
                key_name = destination.path
            else:
                key_name = source

            mimetype, _ = mimetypes.guess_type(source)
            if mimetype is None:
                raise Exception("Failed to guess mimetype")

            self.stdout.write(f"Upload {source_raw} ->  {destination_raw}")
            bucket.upload_file(
                source, key_name, ExtraArgs={"ContentType": mimetype}
            )

        if not is_destination_remote and is_source_remote:
            # Local destination and remote source.
            # Pulling down

            cloud_output = utils.get_cloud_output_from_cdev_name(source.component_name, source.cdev_bucket_name)
            bucket_name = cloud_output.get('bucket_name')
            

            bucket = s3.Bucket(bucket_name)

            self.stdout.write(f"Download {source_raw} -> {destination_raw} ")
            bucket.download_file(source.path, destination)


        if is_destination_remote and is_source_remote:
            # Local destination and remote source.
            # Pulling down

            source_cloud_output = utils.get_cloud_output_from_cdev_name(source.component_name, source.cdev_bucket_name)
            source_bucket_name = source_cloud_output.get('bucket_name')

            destination_cloud_output = utils.get_cloud_output_from_cdev_name(destination.component_name, destination.cdev_bucket_name)
            destination_bucket_name = destination_cloud_output.get('bucket_name')
            

            destination_bucket = s3.Bucket(destination_bucket_name)

            destination_key = destination.path if destination.path else source.path

            self.stdout.write(f"Copy {source_raw} -> {destination_raw}")
            destination_bucket.copy(
                {
                    'Bucket': source_bucket_name,
                    'Key': source.path
                }, 
                destination_key
            )









"""Command for syncing a set of resources to a static site
"""
from argparse import ArgumentParser
import boto3
import os
import mimetypes

from core.constructs.commands import BaseCommand, OutputWrapper
from core.default.resources.simple.static_site import StaticSite, simple_static_site_model
from core.utils.paths import get_full_path_from_workspace_base

from . import utils


class sync_files(BaseCommand):
    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "resource_name", type=str, help="The static site resource you want to sync"
        )
        parser.add_argument(
            "--dir",
            type=str,
            help="override the default content folder of the resource.",
        )
        parser.add_argument(
            "--clear",
            action="store_true",
            help="Clear the existing content of the bucket before syncing the new data.",
        )

    def command(self, *args, **kwargs):
        full_resource_name: str = kwargs.get("resource_name")
        component_name = full_resource_name.split('.')[0]
        static_site_name = full_resource_name.split('.')[1]
        override_directory = kwargs.get("dir")
        clear_bucket = kwargs.get("clear")

        resource: simple_static_site_model = utils.get_resource_from_cdev_name(component_name, static_site_name)
        cloud_output = utils.get_cloud_output_from_cdev_name(component_name, static_site_name)

        if not override_directory:
            final_dir = get_full_path_from_workspace_base(resource.content_folder)

        else:
            final_dir = get_full_path_from_workspace_base(override_directory)

        bucket_name = cloud_output.get("bucket_name")
        

        s3 = boto3.resource("s3")
        bucket = s3.Bucket(bucket_name)

        if clear_bucket:
            bucket.object_versions.delete()

        for subdir, dirs, files in os.walk(final_dir):
            for file in files:
                full_path = os.path.join(subdir, file)

                key_name = os.path.relpath(full_path, final_dir)

                mimetype, _ = mimetypes.guess_type(full_path)
                if mimetype is None:
                    raise Exception("Failed to guess mimetype")
                print(f"Uploading file -> {full_path}")
                bucket.upload_file(
                    full_path, key_name, ExtraArgs={"ContentType": mimetype}
                )



class CdevError(Exception):
    pass



class CouldNotParseFileError(CdevError):
    def __init__(self, error):
        self.error = error


class CdevFileNotFoundError(FileNotFoundError):
    def __init__(self, message):
        self.message = message


class InvalidParamError(CdevError):
    # This error should be thrown when anything is passed a wrong param
    def __init__(self, message):
        self.message = message

class InvalidDataError(CdevError):
    # This error should be thrown when there is a problem with the integrity of data passed into a system
    def __init__(self, message):
        self.message = message



"""
Utilities to help standardize how the framework dynamically imports code
"""

import sys
import importlib
from types import ModuleType
import sys
from typing import TextIO
import inspect


def import_module(module_name: str, denote_output: bool = False, override_stdout: bool = True) -> ModuleType:
    """Helper function for dynamically loading python modules.
    
    This function should be used whenever a python module needs to be dynamically loaded within the framework. 
    
    For more information about python modules see https://docs.python.org/3/tutorial/modules.html. Any changes
    to the PYTHONPATH needed to find the module should be done before this function is called.
    
    Arguments:
        module_name (str): The python module to be loaded. Note that this is not a file path. It is the name of a module that can be loaded.

    Raises:
        Exception
    """

    if denote_output:
        print(f"--------- OUTPUT FROM MODULE {module_name} ----------")

    # Sometimes the module is already loaded so just reload it to capture any changes
    # Importing the initialization file should cause it to modify the state of the Workspace however is needed

    if override_stdout:
        sys.stdout = override_sys_out

    if sys.modules.get(module_name):
        rv = importlib.reload(sys.modules.get(module_name))
    else:
        rv = importlib.import_module(module_name)

    if override_stdout:
        sys.stdout = sys.__stdout__

    if denote_output:
        print(f"---------------------------------------------------")

    return rv


def import_class(module_name: str, class_name: str) -> type:
    """Helper function for loading a specific class from a module

    Import the given module name and then search through all the objects in the module
    for a class with a matching name. Return the `class` if found.

    Args:
        module_name (str): module name
        class_name (str): class name

    Returns:
        type: the class
    """
    mod = import_module(module_name)

    for item_name in dir(mod):

        if not item_name == class_name:
            continue

        potential_obj = getattr(mod, item_name)
        if inspect.isclass(potential_obj):
            return potential_obj

        
    raise Exception

class override_sys_out(TextIO):
   
    encoding = sys.__stdout__.encoding 

    def write(s: str):
        if len(s) > 0 and not s == "\n":
            sys.__stdout__.write(f"> {s}\n")



from dataclasses import dataclass
import re
from tokenize import group

from core.constructs.resource import ResourceModel
from core.constructs.workspace import Workspace

RUUID = "cdev::simple::bucket"

def get_cloud_output_from_cdev_name(component_name: str, cdev_name: str) -> str:
    try:
        ws = Workspace.instance()


        cloud_output = ws.get_backend().get_cloud_output_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_name
        )

        return cloud_output
    except Exception as e:
        print(f"Could not find resource {component_name}:{RUUID}:{cdev_name}")
        print(e)
        return None



def get_resource_from_cdev_name(component_name: str, cdev_name: str) -> ResourceModel:
    try:
        ws = Workspace.instance()


        resource = ws.get_backend().get_resource_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_name
        )

        return resource
    except Exception as e:
        print(f"Could not find resource {component_name}:{RUUID}:{cdev_name}")
        print(e)
        return None


remote_name_regex = "bucket://([a-z,_]+).([a-z,_]+)/?(\S+)?"
compiled_regex = re.compile(remote_name_regex)

@dataclass
class remote_location:
    component_name: str
    cdev_bucket_name: str
    path: str

def is_valid_remote(name: str) -> bool:
    return True if compiled_regex.match(name) else False


def parse_remote_location(name: str) -> remote_location:
    match = compiled_regex.match(name)

    if not match:
        raise Exception("provided name {name} does not match regex for a remote bucket object")

    return remote_location(
        component_name=match.group(1),
        cdev_bucket_name=match.group(2),
        path=match.group(3)
    )






"""Set of constructs for making No SQL DB Tables
"""
from enum import Enum
from typing import Any, FrozenSet, List

from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs, PermissionsAvailableMixin
from core.constructs.cloud_output import  Cloud_Output_Str, OutputType
from core.constructs.types import cdev_str_model
from core.constructs.models import ImmutableModel

from core.default.resources.simple.events import Event, event_model
from core.default.resources.simple.iam import Permission

from core.utils import hasher

RUUID = "cdev::simple::table"


#################
##### Events
#################
#TODO: Add more documentation on the json schema of the events
class stream_type(str, Enum):
    """Type of streams for a table."""

    KEYS_ONLY = "KEYS_ONLY"
    """Only the key attributes of the modified item."""
    NEW_IMAGE = "NEW_IMAGE"
    """The entire item, as it appears after it was modified."""
    OLD_IMAGE = "OLD_IMAGE"
    """The entire item, as it appeared before it was modified."""
    NEW_AND_OLD_IMAGES = "NEW_AND_OLD_IMAGES"
    """Both the new and the old item images of the item."""


class stream_event_model(event_model):
    """Model to represent a stream event

    Args:
        table_name (str): Cdev name of the API this route is apart of
        view_type (stream_type): Type of eventthe stream produces
        batch_size (int): Size of event batches
    """
    table_name: cdev_str_model    
    view_type: stream_type
    batch_size: int


class StreamEvent(Event):
    """Construct for representing the Stream of a `Table`"""

    def __init__(self, table_name: str, view_type: stream_type, batch_size: int) -> None:
        """
        Args:
            table_name (str): Cdev name of the API this route is apart of
            view_type (stream_type): Type of eventthe stream produces
            batch_size (int): Size of event batches
        """
        
        self.cdev_table_name = table_name
        
        self.table_name= Cloud_Output_Str(
                name=table_name, 
                ruuid=RUUID, 
                key='table_name',
                type=OutputType.RESOURCE 
            )
            
        self.view_type=view_type
        self.batch_size=batch_size


    def hash(self) -> str:
        return hasher.hash_list([
            self.cdev_table_name,
            self.view_type,
            self.batch_size
        ])
        
    def render(self) -> stream_event_model:
        return stream_event_model(
            originating_resource_name=self.cdev_table_name,
            originating_resource_type=RUUID,
            hash=self.hash(),
            batch_size=self.batch_size,
            view_type=self.view_type.value,
            table_name=self.table_name.render()
        )


#####################
###### Permission
######################
class TablePermissions:
    """Permissions to provide to other resources to access a `Table`"""

    def __init__(self, resource_name: str) -> None:

        self.READ_TABLE = Permission(
            actions=[
                "dynamodb:GetItem",
                "dynamodb:BatchGetItem",
                "dynamodb:Scan",
                "dynamodb:Query",
                "dynamodb:ConditionCheckItem",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to read data from the Table"""

        self.WRITE_TABLE = Permission(
            actions=[
                "dynamodb:BatchGetItem",
                "dynamodb:BatchWriteItem",
                "dynamodb:ConditionCheckItem",
                "dynamodb:PutItem",
                "dynamodb:DescribeTable",
                "dynamodb:DeleteItem",
                "dynamodb:GetItem",
                "dynamodb:Scan",
                "dynamodb:Query",
                "dynamodb:UpdateItem",
                "dynamodb:DescribeLimits",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to write data to the Table"""

        self.READ_AND_WRITE_TABLE = Permission(
            actions=[
                "dynamodb:BatchGetItem",
                "dynamodb:BatchWriteItem",
                "dynamodb:ConditionCheckItem",
                "dynamodb:DeleteItem",
                "dynamodb:DescribeLimits",
                "dynamodb:DescribeTable",
                "dynamodb:GetItem",
                "dynamodb:PutItem",
                "dynamodb:Query",
                "dynamodb:Scan",
                "dynamodb:UpdateItem",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permissions to read and write data to and from the Table"""

        self.READ_STREAM = Permission(
            actions=[
                "dynamodb:DescribeStream",
                "dynamodb:GetRecords",
                "dynamodb:GetShardIterator",
                "dynamodb:ListShards",
                "dynamodb:ListStreams",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE).join(["", "/stream/*"]),
            effect="Allow",
        )
        """Permissions to receive and process events from the Table Stream"""

##############
##### Output
##############
class TableOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after a `Table` has been deployed."""
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)

    @property
    def table_name(self) -> Cloud_Output_Str:
        """The name of the generated table"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='table_name',
            type=self.OUTPUT_TYPE
        )

    @table_name.setter
    def table_name(self, value: Any):
        raise Exception


###############
##### Table
###############
class attribute_type(str, Enum):
    """Attributes of a table.

    These values will be used by the table to do type checks on data for the defined attribute.

    visit https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Client.create_table for more details
    """

    S = "S"
    """String"""
    N = "N"
    """Number"""
    B = "B"
    """Binary"""


class key_type(str, Enum):
    """Type of key for a defined key on a table.
    
    These value will be used by the primary key to determine how the data is stored in the table.
    visit https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/HowItWorks.CoreComponents.html#HowItWorks.CoreComponents.PrimaryKey for more details.
    """

    HASH = "HASH"
    """Partion Key"""
    RANGE = "RANGE"
    """Sort key"""


class attribute_definition_model(ImmutableModel):
    """Model representing an attribute of a `Table`

    Args:
        attribute_name (str)
        attribute_type (`attribute_type`)
    """
    attribute_name: str 
    attribute_type: attribute_type

    
class AttributeDefinition:
    """Construct representing an attribute of a `Table`"""

    def __init__(self, name:str, type: attribute_type) -> None:
        """Create an attribute that will be part of a `Table`
        
        Args:
            name (str): Name of the attribute
            type (attribute_type): Type of value the attribute stores
        """
        self.name = name
        self.type = type


    def render(self) -> attribute_definition_model:
        return attribute_definition_model(
            attribute_name=self.name,
            attribute_type=self.type
        )


class key_definition_model(ImmutableModel):
    """Model representing a primary key of a `Table`

    Args:
        attribute_name (str)
        key_type (`key_type`)
    """
    attribute_name: str 
    key_type: key_type


class KeyDefinition:
    """Construct representing a primary key of a `Table`"""

    def __init__(self, name:str, type: key_type) -> None:
        """
        Args:
            name (str): Name of the attribute
            type (key_type): Type of value the attribute stores

        Note that the `name` property must match a `name` of a given `AttributeDefinition` on the created
        table. 

        The keys on a table will define the optimal way of retrieving data from the `Table`. 


        """
        self.name = name
        self.type = type


    def render(self) -> key_definition_model:
        return key_definition_model(
            attribute_name=self.name,
            key_type=self.type
        )


class simple_table_model(ResourceModel):
    """Model representing a `Table`

    Args:
        ResourceModel ([type]): [description]
    """
    attributes: FrozenSet[attribute_definition_model]
    keys: FrozenSet[key_definition_model]


class Table(PermissionsAvailableMixin, Resource):
    """Create a NoSql Table (DynamoDB)"""

    @update_hash
    def __init__(
        self,
        cdev_name: str,
        attributes: List[AttributeDefinition],
        keys: List[KeyDefinition],
        nonce: str = "",
    ) -> None:
        """
        Args:
            cdev_name (str): Name for the resource.
            attributes (List[AttributeDefinition]): List of Attributes on the Table
            keys (List[KeyDefinition]): List of Key Definitions that make up the primary keys
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.


        <a href='https://code.tutsplus.com/tutorials/a-beginners-guide-to-http-and-rest--net-16340'>More information on how to use a DynamoDB Table</a>

        <a href="/docs/examples/table"> Examples on how to use in Cdev Framework</a>

        <a href="https://docs.aws.amazon.com/dynamodb/index.html"> Documentation on Deployed Resource in the Cloud</a>

        <a href="https://aws.amazon.com/dynamodb/pricing/"> Details on pricing</a>
        """
        super().__init__(cdev_name, RUUID, nonce)

        self._attributes = attributes
        self._keys = keys
        self._stream = None

        self._available_permissions: TablePermissions = TablePermissions(cdev_name)
        self._output = TableOutput(cdev_name)

    @property
    def output(self) -> TableOutput:
        """Output generated by the Cloud when this resource is deployed."""
        return self._output

    @property
    def available_permissions(self) -> TablePermissions:
        """Permissions that can be granted to other resources to access this `Table`"""
        return self._available_permissions

    @property
    def attributes(self) -> List[AttributeDefinition]:
        """Current Attributes"""
        return self._attributes

    @attributes.setter
    @update_hash
    def attributes(self, value: List[AttributeDefinition]):
        self._attributes = value

    @property
    def keys(self) -> List[KeyDefinition]:
        """Current Primary Key Definitions"""
        return self._keys

    @keys.setter
    @update_hash
    def keys(self, value: List[KeyDefinition]):
        self._keys = value

    def create_stream(
        self, view_type: stream_type, batch_size: int = 100
    ) -> StreamEvent:
        """Create a `StreamEvent` for this table

        Args:
            view_type (stream_type): Type of events to process from the stream.
            batch_size (int, optional): Number events to batch into an Event. Defaults to 100.

        Raises:
            Exception: If this Table already has a Stream, throw exception. Should use `get_stream()`

        Returns:
            StreamEvent: The created `Stream_Event`
        """
        if self._stream:
            raise Exception(f"Already created stream on this table. Use `get_stream()` to get the current stream.")

        event = StreamEvent(
            table_name=self.name,
            view_type=view_type,
            batch_size=batch_size
        )

        self._stream = event
        return event

    def get_stream(self) -> StreamEvent:
        """Get the `StreamEvent` for this `Table`

        Raises:
            Exception: If this Table has not created a stream, throw exception

        Returns:
            StreamEvent: The `StreamEvent` for this Table
        """
        if not self._stream:
            raise Exception("Stream has not been created. Create a stream for this table using the `create_stream` function.")

        return self._stream


    def _check_attributes_and_keys(
        self
    ) -> None:
        """
        Check key constraints based on https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Client.create_table
        """
        if len(self.keys) > 2:
            raise Exception("Only two primary keys can be used")

        primary_key = self.keys[0]

        if not primary_key:
            raise Exception("No Hash key provided")

        if not primary_key.type == key_type.HASH:
            raise Exception("First key is not Hash key")

        if not primary_key.name in set(
            [x.name for x in self.attributes]
        ):
            raise Exception(
                f"Hash key 'AttributeName' ({primary_key.name}) not defined in attributes",
            )

        if len(self.keys) == 1:
            return 

        range_key = self.keys[1]

        if not range_key.type == key_type.RANGE:
            raise Exception("Second key is not a Range key")

        if not range_key.name in set(
            [x.name for x in self.attributes]
        ):
            raise Exception(
                f"Range key 'AttributeName' ({range_key.name}) not defined in attributes",
            )

    def compute_hash(self):
        self._hash = hasher.hash_list(
            [
                [x.render() for x in self.attributes], 
                [x.render() for x in self.keys], 
                self.nonce
            ]
        )

    def render(self) -> simple_table_model:
        self._check_attributes_and_keys()
        
        return simple_table_model(
            ruuid=RUUID,
            name=self.name,
            hash=self.hash,
            attributes=[x.render() for x in self.attributes],
            keys=[x.render() for x in self.keys],
        )



import json
from typing import FrozenSet, List, Tuple, Union, Dict
from uuid import uuid4

from core.default.resources.simple.iam import permission_model, permission_arn_model

from .. import aws_client 


def create_role_with_permissions(
    role_name: str,
    permissions: FrozenSet[Union[permission_model, permission_arn_model]],
    assume_role_policy: str
) -> Tuple[str, List[Dict]]:
    """
    This function creates a new IAM role and policies such that the function will have correct access to resources.

    Arguments:
        role_name: Name of the role
        permissions [List[Union[permission_model, permission_arn_model]]]: The permission to assign to the role
    
    Returns:
        str: The arn of the role created
        List[Tuple[str,bool]]: List of Tuple of permission arn and bool to denote if the permission should be destroyed on delete
    
    """

    role_arn = _create_role(role_name, assume_role_policy)
    permission_info: List[Tuple[str, bool]] = []


    for permission in permissions:
        permission_info.append(add_policy(role_name, permission))


    basic_execution_permission_arn = permission_arn_model(
        arn="arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole",
        hash="arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole"
    )

    permission_info.append(
        add_policy(role_name, basic_execution_permission_arn)
    )

    return (role_arn, permission_info)


def delete_role_and_permissions(
    role_name: str, permission_arns: List[Dict]
):
    """
    Delete all permissions and the associated role
    """
    for info in permission_arns:
        delete_policy(role_name, info)

    aws_client.run_client_function("iam", "delete_role", {"RoleName": role_name})




def delete_policy(role_name: str, permission_info: Dict):
    """
    Delete a policy
    """
    detach_policy(role_name, permission_info.get('arn'))

    if permission_info.get('was_created'):
        aws_client.run_client_function(
            "iam", "delete_policy", {"PolicyArn": permission_info.get('arn')}
        )


def add_policy(
    role_name: str,
    permission: Union[permission_model, permission_arn_model]
) -> Dict:
    """
    Creates a policy if needed and then adds the policy to the given role
    """
    if isinstance(permission, permission_arn_model):
        # Already deployed policy so just append arn
        returned_permission_arn = permission.arn
        was_created = False
        
    else:
        returned_permission_arn = create_policy(permission)
        was_created = True
        
    _attach_policy_to_arn(role_name, returned_permission_arn)

    return {
        "hash": permission.hash,
        "arn": returned_permission_arn,
        "was_created": was_created
    }

def create_policy(permission: permission_model) -> str:
    """
    Creates the policy and returns the arn
    """

    policy = {
        "Version": "2012-10-17",
    }

    statement = {"Effect": permission.effect, "Action": list(permission.actions)}

    statement["Resource"] = permission.cloud_id
    policy["Statement"] = [statement]

    permission_name = str(uuid4())

    rv = aws_client.run_client_function(
        "iam",
        "create_policy",
        {
            "PolicyName": permission_name, 
            "PolicyDocument": json.dumps(policy)
        },
    )

    return rv.get("Policy").get("Arn")


def detach_policy(role_name: str, permission_arn: str):
    aws_client.run_client_function(
        "iam",
        "detach_role_policy",
        {
            "RoleName": role_name, 
            "PolicyArn": permission_arn
        },
    )


def _create_role(name: str, assume_role_policy: str) -> str:
    """
    Creates the role and returns the arn
    """
    
    rv = aws_client.run_client_function(
        "iam",
        "create_role",
        {
            "RoleName": name,
            "AssumeRolePolicyDocument": assume_role_policy,
        },
    )

    return rv.get("Role").get("Arn")


def _attach_policy_to_arn(role_arn: str, policy_arn: str) -> bool:   
    aws_client.run_client_function(
        "iam", 
        "attach_role_policy", 
        {
            "RoleName": role_arn, 
            "PolicyArn": policy_arn
        }
    )







from core.utils.exceptions import Cdev_Error, Cdev_Warning
from core.utils.fs_manager import finder


from core.constructs.components import ComponentModel, Component
from core.utils import hasher


class Cdev_FileSystem_Component(Component):
    """
    A simple implementation of a Cdev Component that uses a folder on the file system to render the desired resources.
    This component uses provided libraries in Cdev to generate resources. Using this component, you can create serverless
    functions using the provided Cdev annotations that also provides functionality to parse the desired folder out of the
    file to make it more efficient.
    """

    def __init__(self, fp, name):
        super().__init__(name)
        self.fp = fp

    def render(self) -> ComponentModel:
        """
        Render this component based on the information in the files at the provided folder path
        """
        resources_sorted, references_sorted = finder.parse_folder(
            self.fp
        )
        total_component_hash = hasher.hash_list(
            [x.hash for x in resources_sorted] + [x.hash for x in references_sorted]
        )

        
        rv = ComponentModel(
            **{
                "resources": list(resources_sorted),
                "hash": total_component_hash,
                "name": self.get_name(),
            }
        )

        return rv



"""Utilities for creating a deploying a set of changes 

"""

from core.constructs.output_manager import OutputManager
from ..constructs.workspace import Workspace, Workspace_State
from rich.prompt import Confirm
from .execute_frontend import execute_frontend

import networkx as nx


def execute_deployment_cli(args):

    WORKSPACE = Workspace.instance()

    execute_deployment(WORKSPACE, OutputManager())



def execute_deployment(workspace: Workspace, output: OutputManager):
    unsorted_differences = execute_frontend(workspace, output)

    if not unsorted_differences:
        return

    differences_structured = workspace.sort_differences(unsorted_differences)

    print("")
    do_deployment = Confirm.ask("Do you want to deploy differences?")

    if not do_deployment:
        return

    workspace.set_state(Workspace_State.EXECUTING_BACKEND)

    workspace.deploy_differences(differences_structured)


    

    for tag, cloud_output in workspace.render_outputs():
        print(f"{tag} -> {cloud_output}")



from argparse import ArgumentParser
import json
import os
from tempfile import TemporaryFile
from typing import Dict, List

from core.constructs.commands import BaseCommand, OutputWrapper
from core.default.resources.simple.table import Table, simple_table_model
from core.utils.paths import get_full_path_from_workspace_base

from core.default.mappers import aws_client


from . import utils 


RUUID = "cdev::simple::table"


class put_object(BaseCommand):
    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "resource_name", type=str, help="The resource you want to sync data to"
        )
        parser.add_argument(
            "--file", type=str, help="The location of the json file containing data"
        )
        parser.add_argument(
            "--data", type=str, help="The json data you want to put in the db"
        )

    def command(self, *args, **kwargs):
        full_resource_name = kwargs.get("resource_name")
        component_name = full_resource_name.split('.')[0]
        table_resource_name = full_resource_name.split('.')[1]

        cloud_output = utils.get_cloud_output_from_cdev_name(component_name, table_resource_name)
        resource: simple_table_model = utils.get_resource_from_cdev_name(component_name, table_resource_name)

        table_cloud_name = cloud_output.get('table_name')

        data_string = kwargs.get("data")
        data_file = kwargs.get("file")

        if data_string and data_file:
            raise Exception(f"Can not provide both --data and --file arguments")

        if not (data_string or data_file):
            raise Exception(f"Must provide either --data or --file arguments")



        if data_string:
            try:
                data = json.loads(data_string)
                    
            except Exception as e:
                self.stderr.write("Data provided was not a valid json string")
                return


        if data_file:
            if not os.path.isfile(data_file):
                raise Exception(f"{data_file} is not a valid data file")

            try:
                with open(data_file) as fp:
                    data = json.load(fp)
            
            except Exception as e:
                print(e)
                raise e


            if not 'items' in data:
                raise Exception(f"Loaded data from {data_file} does not contain the 'items' key")


        attributes_dict = {
            x.get("attribute_name"): x.get("attribute_type") for x in resource.attributes
        }


        for datum in data.get('items'):
            is_data_valid, msg = utils.validate_data(datum, attributes_dict, resource.keys)

            if not is_data_valid:
                self.stderr.write(msg)
                return

        try:
            
            translated_data = [utils.recursive_translate_data(x) for x in data.get('items')]
        except:
            self.stderr.write("Could not translate data to dynamodb put item form")

        try:
            for datum in translated_data:
                aws_client.run_client_function(
                    "dynamodb",
                    "put_item",
                    {
                        "TableName": table_cloud_name, 
                        "Item": datum
                    },
                )

                self.stdout.write(f"Wrote {datum} to {table_resource_name}")
        except Exception as e:
            self.stderr.write(f"Could not write data to table {table_resource_name}")
            return

        



from cdev.constructs.project import Project
from cdev.cli.logger import set_global_logger_from_cli

from core.commands.deploy_differences import execute_deployment
from cdev.default.output_manager import CdevOutputManager


def develop_command_cli(args):
    config = args[0]
    set_global_logger_from_cli(config.loglevel)
    develop_command(args)


def develop_command(args):

    output_manager = CdevOutputManager()
    myProject = Project.instance()

    print(f"Live Development")



"""import os
import sys
from time import sleep
from typing import Dict, List, Tuple

from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.live import Live
from rich.text import Text

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from cdev.constructs import Cdev_Project

from cdev.utils import hasher

import threading
    
from ..utils import project
from ..frontend import executer as frontend_executer
from ..backend import executer as backend_executer
from ..backend import resource_state_manager, cloud_mapper_manager
    
from cdev import output as cdev_output
from cdev.settings import set_setting

from . import deploy

STD_OUT_HISTORY_BUFFER = []
history = []
Program_Executing = True

CLOUD_OUTPUT_BUFFER = "CLOUD_OUTPUT"

CURRENT_CONTEXT = {
    "is_deploy_running": False
}


def make_develop_layout() -> Layout:
    layout = Layout("tmp")
    layout.split(
        
        Layout(name="stdout"),
        Layout(name="cloud_output"), 
        Layout(name="commands"), 
     
    )

    layout['stdout'].ratio = 80
    layout['cloud_output'].ratio = 20
    layout['commands'].ratio = 1
    #layout['hidden'].visible = False

    layout['commands'].update("[blink]----- Command ------[blink]")
    layout['cloud_output'].update(Panel("", title="Cloud Output"))
    return layout



LAYOUT = make_develop_layout()
LIVE_OBJECT = Live(LAYOUT, auto_refresh=False, transient=True)

def develop(args):
    if not args.simple:
        set_setting("CAPTURE_OUTPUT", True)

    run_enhanced_local_development_environment(args)

def run_enhanced_local_development_environment(args):
    patterns = ["./src/*"]
    ignore_patterns = ["*/__pycache__/*"]
    ignore_directories = True
    case_sensitive = True
    my_event_handler = PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)


    path = os.getcwd()
    go_recursively = True
    my_observer = Observer()

    if args.simple:
        #my_event_handler.on_created = file_change_handler_simple
        #my_event_handler.on_deleted = file_change_handler_simple
        my_event_handler.on_modified = file_change_handler_simple
        #my_event_handler.on_moved = file_change_handler_simple

        my_observer.schedule(my_event_handler, path, recursive=go_recursively)

    else:
        my_event_handler.on_created = file_change_handler
        my_event_handler.on_deleted = file_change_handler
        my_event_handler.on_modified = file_change_handler
        my_event_handler.on_moved = file_change_handler

        my_observer.schedule(my_event_handler, path, recursive=go_recursively)
    

    
    cdev_output.create_buffer(CLOUD_OUTPUT_BUFFER)
    refresh_local_output({"buffer_name": CLOUD_OUTPUT_BUFFER, "reinitialize_project": True})
    my_observer.start()
    cdev_output.print(f"")
    cdev_output.print(f"[blink] *** waiting for changes ***[/blink]")
    
    handle_std_in_thread = threading.Thread(target=handle_std_input, daemon=True)
    try:
        
        if args.simple:
            # IF this is a simple develop session don't create live panels
            while True:
                pass

        
        with LIVE_OBJECT as l:
            refresh_output_buffer()
            
            handle_std_in_thread.start()
            last_stdout_hash = 0
            while True:
                messages, start_line_no, messages_hash = cdev_output.get_messages_from_buffer(-25,None)
                
                modified_messages = [f"({start_line_no+i}) {x}" for i,x in enumerate(messages,0)]

                if messages_hash == last_stdout_hash:
                    pass
                else:
                    
                    messages_as_string = "\n".join(modified_messages)
                    last_stdout_hash = messages_hash
                    LAYOUT['stdout'].update(Panel(messages_as_string, title="STD OUT"))
                    update_screen()
                    
                    sleep(.1)


    except KeyboardInterrupt:
        print("Development environment closed1")
        my_observer.stop()
        my_observer.join()

        if not args.simple:
            handle_std_in_thread.join()
        exit(0)


def refresh_output_buffer():
    cdev_output.clear_buffer(CLOUD_OUTPUT_BUFFER)
    refresh_local_output({"buffer_name": CLOUD_OUTPUT_BUFFER, "reinitialize_project": False})
    cloud_outputs,_,_ = cdev_output.get_messages_from_buffer(0,10, CLOUD_OUTPUT_BUFFER)
    LAYOUT['cloud_output'].update(Panel("\n".join(cloud_outputs), title="Cloud Output"))


def refresh_output_simple():
    refresh_local_output({"reinitialize_project": False})



def handle_std_input():
    while True:
        command = []
        for line in sys.stdin.readline():
            # run command
            command.append(line)

        cdev_output.print("".join(command)[:-1])
        refresh_output_buffer()
        update_screen()


def file_change_handler(args):
    cdev_output.print(f"File Change Detected. Starting Deploy Process")
    deploy.local_deploy_command({})
    cdev_output.print(f"[blink] *** waiting for changes ***[/blink]")
    refresh_output_buffer()
    update_screen()

def file_change_handler_simple(args):
    CDEV_PROJECT = Cdev_Project()
    CDEV_PROJECT.clear_previous_state()
    cdev_output.print(f"File Change Detected. Starting Deploy Process")
    did_deploy = deploy.local_deploy_command({})
    if did_deploy:
        refresh_output_simple()
    cdev_output.print(f"[blink] *** waiting for changes ***[/blink]")
    
    update_screen()




def update_screen():
    LIVE_OBJECT.update(LAYOUT, refresh=True)


def add_line_to_history(line):
    history.append(line)

    

def get_output_buffer() -> Tuple[str, str]:
    if not history:
        return None

    lines = '\n'.join(history)
    hash_val = hasher.hash_string(lines)

    #history.clear()
    return lines, hash_val


def refresh_local_output(args: Dict):  

    if not "buffer_name" in args:
        write_to_buffer = False
    else:
        write_to_buffer = True

    if args.get("reinitialize_project"):
        project.initialize_project()
        frontend_executer.execute_frontend()


    PROJECT = project.Cdev_Project()
    
    desired_outputs = PROJECT.get_outputs()

    rendered_outputs = []

    for label, output in desired_outputs.items(): 
        
        identifier = output.resource.split("::")[-1]

        if output.transformer:
            rendered_value = cloud_mapper_manager.get_output_value_by_hash(identifier, output.key, transformer=output.get("transformer"))
        else:
            rendered_value = cloud_mapper_manager.get_output_value_by_hash(identifier, output.key)

        rendered_outputs.append(f"[magenta]{label}[/magenta] -> [green]{rendered_value}[/green]")

    if not write_to_buffer:
        cdev_output.print("---Current Output---")
    else:
        cdev_output.add_message_to_buffer(args.get("buffer_name"), "---Current Output---")

    for rendered_output in rendered_outputs:
        if not write_to_buffer:
            cdev_output.print(rendered_output)
        else:
            cdev_output.add_message_to_buffer(args.get("buffer_name"), str(rendered_output))

    


"""



import json
from pydantic.types import FilePath
from typing import Dict, List, Callable, Any, TypeVar, Tuple


from cdev.constructs.project import Project, Project_State, project_info, wrap_phases
from cdev.default.environment import local_environment
from core.constructs.backend import Backend, load_backend

from core.constructs.mapper import CloudMapper
from core.constructs.components import Component
from core.constructs.workspace import Workspace_State

from core.constructs.workspace import Workspace_Info
from core.constructs.settings import Settings_Info, Settings
from core.constructs.cloud_output import Cloud_Output

from core.utils import file_manager

from ..constructs.environment import environment_info, Environment

WORKSPACE_STATE_TO_PROJECT_STATE = {
    Workspace_State.UNINITIALIZED: Project_State.UNINITIALIZED,
    Workspace_State.INITIALIZING: Project_State.INITIALIZING,
    Workspace_State.INITIALIZED: Project_State.INITIALIZED,
}


class local_project(Project):
    """
    An implementation of the Project API that works for simple local development.

    Arguments:
        project_info_location (FilePath): Path the configuration json file

    """

    _instance = None

    _current_environment: Environment = None
    _project_info_location: FilePath = None

    _central_state: project_info
    _backend: Backend = None
    _project_state: Project_State = None
    _project_name: str = None

    def __new__(cls, project_info_location: FilePath):
        if cls._instance is None:
            cls._instance = super(Project, cls).__new__(cls)
            # Put any initialization here.
            cls._instance._project_info_location = project_info_location

            cls._instance._load_state()

            cls._instance.set_name(cls._instance._central_state.project_name)

            cls._instance._backend = load_backend(
                cls._instance._central_state.backend_info
            )

            # Load the backend
            cls._instance._current_environment = None
            
            cls._instance.set_state(Project_State.UNINITIALIZED)

            Project.set_global_instance(cls._instance)


        return cls._instance

    @classmethod
    def terminate_singleton(cls):
        cls._instance = None

    def initialize_project(self):
        self.set_state(Project_State.INITIALIZING)
        current_env = self.get_current_environment()
        
        if current_env:
            current_env.initialize_environment()
        self.set_state(Project_State.INITIALIZED)

    def terminate_project(self):
        Project.remove_global_instance(self)


    def set_name(self, name: str):
        """
        Set the name of this Project

        Args:
            name (str): name of the Project
        """
        self._project_name = name


    def get_name(self):
        """
        Get the name of this Project

        Returns:
            name (str)
        """
        return self._project_name

    def get_state(self) -> Project_State:
        return self._project_state

    def set_state(self, new_state: Project_State):
        self._project_state = new_state

    # Note that the class methods should not include doc strings so that they inherit the doc string of the 
    # parent class. 

    @wrap_phases([Project_State.INITIALIZED, Project_State.UNINITIALIZED])
    def create_environment(self, environment_name: str, settings_files: Dict[str,str] = None):
        self._load_state()
        resource_state_id = self._backend.create_resource_state(environment_name)

        workspace_config = {"backend_configuration": self._central_state.backend_info}

        workspace_config["resource_state_uuid"] = resource_state_id
        workspace_config["initialization_module"] = "src.cdev_project"

            
        settings = Settings_Info(
            base_class = "core.constructs.settings.Settings"
        ) if not settings_files else Settings_Info(
            base_class = "core.constructs.settings.Settings",
            user_setting_module=settings_files.get('user_setting_module'),
            secret_dir=settings_files.get('secret_dir'),
        )

        self._central_state.environments.append(
            environment_info(
                environment_name,
                Workspace_Info(
                    "core.default.workspace", 
                    "local_workspace", 
                    settings,
                    workspace_config
                ),
            )
        )

        self._write_state()

    @wrap_phases([Project_State.INITIALIZED, Project_State.UNINITIALIZED])
    def destroy_environment(self, environment_name: str) -> None:
        self._load_state()

        self._central_state.environments = [
            x for x in self._central_state.environments if x.name != environment_name
        ]

        self._write_state()

    @wrap_phases([Project_State.INITIALIZED, Project_State.UNINITIALIZED])
    def get_all_environment_names(self) -> List[str]:
        self._load_state()

        return [x.name for x in self._central_state.environments]

    @wrap_phases([Project_State.INITIALIZED, Project_State.UNINITIALIZED])
    def set_current_environment(self, environment_name: str):
        self._load_state()

        if not environment_name in self.get_all_environment_names():
            raise Exception

        self._central_state.current_environment = environment_name

        self._write_state()

    def get_environment(self, environment_name: str) -> Environment:
        self._load_state()

        environment_info = next(
            x for x in self._central_state.environments if x.name == environment_name
        )

        return local_environment(environment_info)

    def get_current_environment_name(self) -> str:
        self._load_state()

        return self._central_state.current_environment

    def get_current_environment(self) -> Environment:
        self._load_state()

        if not self._central_state.current_environment:
            return None

        return self.get_environment(self._central_state.current_environment)

    def _get_environment_info(self, name: str) -> environment_info:
        self._load_state()

        lookup_dict = {x.name: x for x in self._central_state.environments}

        if not name in lookup_dict:
            raise Exception(f"No environment with name {name}")

        return lookup_dict.get(name)

    ############################
    ##### Settings
    ############################
    @property
    def settings(self) -> Settings:
        return self.get_current_environment().get_workspace().settings

    
    @settings.setter
    def settings(self, value: Settings):
        self.get_current_environment().get_workspace().settings= value


    def get_settings_info(self, environment_name: str =  None) -> Settings_Info:
        if not environment_name:
            environment_name = self.get_current_environment_name()


        self._load_state()
        
        if not environment_name in [x.name for x in self._central_state.environments]:
            raise Exception(f"No environment named {environment_name}")


        return [x for x in self._central_state.environments if x.name == environment_name][0].workspace_info.settings_info
    
    def update_settings_info(self, new_value: Settings_Info, environment_name: str = None):
        if not environment_name:
            environment_name = self.get_current_environment_name()
            
        self._load_state()

        # Remove the old environment
        previous_environment_var = [x for x in self._central_state.environments if x.name == environment_name][0]
        previous_environment_var.workspace_info.settings_info = new_value


        self._central_state.environments = [x for x in self._central_state.environments if not x.name == environment_name]

        self._central_state.environments.append(previous_environment_var)

        self._write_state()

    #######################
    ##### Display Output
    #######################
    def display_output(self, tag: str, output: Cloud_Output):
        """Display the output from a Resource or Reference after a process has completed

        Args:
            tag: A key value to display with the output
            output: The Cloud Output to render
        """
        self.get_current_environment().get_workspace().display_output(tag, output)


    def render_outputs(self) -> List[Tuple[str, Any]]:
        """Render the output associated with the Workspace

        Returns:
            List[Tuple[str, Any]]: The List of outputs with their associated tag
        
        """
        return self.get_current_environment().get_workspace().render_outputs()


    #################
    ##### Mappers
    #################
    @wrap_phases([Workspace_State.INITIALIZING])
    def add_mapper(self, mapper: CloudMapper) -> None:
        ws = self.get_current_environment().get_workspace()
        ws.add_mapper(mapper)

    @wrap_phases([Workspace_State.INITIALIZING])
    def add_mappers(self, mappers: List[CloudMapper]) -> None:
        ws = self.get_current_environment().get_workspace()
        ws.add_mappers(mappers)

    @wrap_phases([Workspace_State.INITIALIZED])
    def get_mappers(self) -> List[CloudMapper]:
        ws = self.get_current_environment().get_workspace()
        return ws.get_mappers()

    @wrap_phases([Workspace_State.INITIALIZED])
    def get_mapper_namespace(self) -> Dict:
        ws = self.get_current_environment().get_workspace()
        return ws.get_mapper_namespace()

    #################
    ##### Commands
    #################
    @wrap_phases([Workspace_State.INITIALIZING])
    def add_command(self, command_location: str):
        ws = self.get_current_environment().get_workspace()
        ws.add_command(command_location)

    @wrap_phases([Workspace_State.INITIALIZING])
    def add_commands(self, command_locations: List[str]):
        ws = self.get_current_environment().get_workspace()
        ws.add_commands(command_locations)

    @wrap_phases([Workspace_State.INITIALIZED])
    def get_commands(self) -> List[str]:
        ws = self.get_current_environment().get_workspace()
        return ws.get_commands()

    #################
    ##### Components
    #################
    @wrap_phases([Workspace_State.INITIALIZING])
    def add_component(self, component: Component):
        ws = self.get_current_environment().get_workspace()
        ws.add_component(component)

    @wrap_phases([Workspace_State.INITIALIZING])
    def add_components(self, components: List[Component]):
        ws = self.get_current_environment().get_workspace()
        ws.add_components(components)

    @wrap_phases([Workspace_State.INITIALIZED])
    def get_components(self) -> List[Component]:
        ws = self.get_current_environment().get_workspace()
        return ws.get_components()

    def _write_state(self):
        file_manager.safe_json_write(
            self._central_state.dict(), self._project_info_location
        )

    def _load_state(self) -> project_info:
        with open(self._project_info_location, "r") as fh:
            self._central_state = project_info(**json.load(fh))



"""Structures for user defined command system

One of the goals of the Core Framework is to provide flexibility such that end users can easily
create custom functionality for their projects. One way of providing this flexibility, is to provide
an easy interface for executing code from within the context of the framework from the CLI. This 
process is heavily modeled and inspired by the `Django Command Framework` 
(https://github.com/django/django/blob/b0ed619303d2fb723330ca9efa3acf23d49f1d19/django/core/management/base.py).

The main primitive is the `BaseCommand` class. By deriving from this class, developers can create code
that is easily accessible and discoverable from the command line. To see how these class are discovered 
and execute see our more in depth documentation on the command system at <link>

"""

import os
from argparse import ArgumentParser, HelpFormatter
import sys
from io import TextIOBase
from rich.console import Console


class CdevCommandError(BaseException):
    """Base exception Class

    Exception class indicating a problem while executing a management
    command.
    If this exception is raised during the execution of a management
    command, it will be caught and turned into a nicely-printed error
    message to the appropriate output stream (i.e., stderr); as a
    result, raising this exception (with a sensible description of the
    error) is the preferred way to indicate that something has gone
    wrong in the execution of a command.
    """

    def __init__(self, *args, returncode=1, **kwargs):
        self.returncode = returncode
        super().__init__(*args, **kwargs)


class CdevHelpFormatter(HelpFormatter):
    """
    This is just a nice feature of the formatter. Thank you Django.

    Customized formatter so that command-specific arguments appear in the
    --help output before arguments common to all commands.
    """

    show_last = {
        "--version",
        "--verbosity",
        "--traceback",
        "--settings",
        "--pythonpath",
        "--no-color",
        "--force-color",
        "--skip-checks",
    }

    def _reordered_actions(self, actions):
        return sorted(
            actions, key=lambda a: set(a.option_strings) & self.show_last != set()
        )

    def add_usage(self, usage, actions, *args, **kwargs):
        super().add_usage(usage, self._reordered_actions(actions), *args, **kwargs)

    def add_arguments(self, actions):
        super().add_arguments(self._reordered_actions(actions))


class OutputWrapper(TextIOBase):
    """
    Wrapper around stdout/stderr
    """

    @property
    def style_func(self):
        return self._style_func

    @style_func.setter
    def style_func(self, style_func):
        if style_func and self.isatty():
            self._style_func = style_func
        else:
            self._style_func = lambda x: x

    def __init__(self, out, ending="\n"):
        self._out = out
        self.style_func = None
        self.ending = ending

    def __getattr__(self, name):
        return getattr(self._out, name)

    def flush(self):
        if hasattr(self._out, "flush"):
            self._out.flush()

    def isatty(self):
        return hasattr(self._out, "isatty") and self._out.isatty()

    def write(self, msg="", style_func=None, ending=None):
        ending = self.ending if ending is None else ending
        if ending and not msg.endswith(ending):
            msg += ending
        style_func = style_func or self.style_func
        self._out.write(style_func(msg))


class ConsoleOutputWrapper(TextIOBase):
    """
    Wrapper around a `rich` console. 
    """

    @property
    def style_func(self):
        return self._style_func

    @style_func.setter
    def style_func(self, style_func):
        if style_func and self.isatty():
            self._style_func = style_func
        else:
            self._style_func = lambda x: x

    def __init__(self, console: Console, ending="\n"):
        self._out = console
        self.style_func = None
        self.ending = ending

    def __getattr__(self, name):
        return getattr(self._out, name)

    def flush(self):
        self._out.clear()

    def isatty(self):
        return self._out.is_terminal

    def write(self, msg: str ="", style_func=None, ending=None):
        ending = self.ending if ending is None else ending
        if ending and not msg.endswith(ending):
            msg += ending
        style_func = style_func or self.style_func
        self._out.print(style_func(msg))



class BaseCommand:
    """Base Class for a user defined command
    
    """

    # Help message for this command
    help = ""

    # If this command requires the project to have been initialized
    requires_initialized_workspace = True

    # If the command should look to substitute args with cloud output values
    substitute_from_output = True

    def __init__(self, stdout: Console = None, stderr: Console = None, no_color=False, force_color=False):
        self.stdout = ConsoleOutputWrapper(stdout) if stdout else OutputWrapper(sys.stdout)
        self.stderr = ConsoleOutputWrapper(stderr) if stderr else OutputWrapper(sys.stderr)

    def create_arg_parser(self, prog_name, subcommand, **kwargs) -> ArgumentParser:
        parser = ArgumentParser(
            prog="%s %s" % (os.path.basename(prog_name), subcommand),
            description=self.help or None,
            formatter_class=CdevHelpFormatter,
            **kwargs
        )

        self.add_arguments(parser)

        return parser

    def add_arguments(self, parser: ArgumentParser):
        """Add the cli arguments for the command
        
        """
        pass

    def run_from_command_line(self, argv):
        """
        Handles input from command line and builds arg parser and uses it to validate input.

        argv -> [program_name, command, *args]

        program_name: represents where to find the command
        command: command to run
        *args: args for the command
        """
        parser = self.create_arg_parser(argv[0], argv[1])

        options = parser.parse_args(argv[2:])

        cmd_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop("args", ())

        try:
            self.command(*args, **cmd_options)
        except CdevCommandError as e:
            self.stderr.write("%s: %s" % (e.__class__.__name__, e))
            sys.exit(e.returncode)

    def command(self, *args, **kwargs):
        """
        The actual logic of the command. Subclasses must implement
        this method.
        """
        raise NotImplementedError(
            "subclasses of BaseCommand must provide a command() method"
        )


class BaseCommandContainer:
    """
    This is used to designate that the directory is a CommandContainer and should be searched for when looking for commands.

    This class should be initialized in the __init__.py file for the folder.

    The directory should contain valid Commands or other CommandContainers
    """

    # Help message for this container
    help = ""

    def __init__(self, stdout: Console = None, stderr: Console = None, no_color=False, force_color=False):
        self.stdout = ConsoleOutputWrapper(stdout) if stdout else OutputWrapper(sys.stdout)
        self.stderr = ConsoleOutputWrapper(stderr) if stderr else OutputWrapper(sys.stderr)

    def display_help_message(self) -> str:
        self.stdout.write(self.help)



# Generated as part of Quick Start project template 

from cdev.resources.simple.api import Api
from cdev.resources.simple.xlambda import simple_function_annotation

from cdev import Project as cdev_project

myProject = cdev_project.instance()

DemoApi = Api("demoapi")

hello_route = DemoApi.route("/hello_world", "GET")

@simple_function_annotation("hello_world_function", events=[hello_route.event()])
def hello_world(event, context):
    print('Hello from inside your Function!')


    return {
        "status_code": 200,
        "message": "Hello Outside World!"
    }



myProject.display_output("Base API URL", DemoApi.output.endpoint)
myProject.display_output("Routes", DemoApi.output.endpoints)


from rich.prompt import Confirm
from typing import List, Tuple

from cdev.constructs.project import Project
from cdev.cli.logger import set_global_logger_from_cli


def environment_cli(args):
    command = args[0]
    parsed_args = vars(args[1])
    
    set_global_logger_from_cli(parsed_args.get("loglevel"))


    if command == "":
        print(
            "You must provide a sub-command. run `cdev environment --help` for more information on available subcommands"
        )
    elif command == "ls":
        list_environments()
    elif command == "get":
        pass
    elif command == "set":
        set_current_environment(parsed_args.get("env"))
    elif command == "create":
        create_environment(parsed_args.get("env"))
    elif command == "settings_information":
        settings_information(parsed_args.get('key'), parsed_args.get('new_value'), parsed_args.get('all'))


def list_environments() -> Tuple[List[str], str]:
    """
    Get the current list of environments and the current environment from the Project object.
    Must be called when the Project is in the UNINITIALIZED phase.
    """

    myProject = Project.instance()

    current_environment_name = myProject.get_current_environment_name()
    all_environment_names = myProject.get_all_environment_names()

    for environment_name in all_environment_names:
        if current_environment_name == environment_name:
            print(f"> {environment_name}")

        else:
            print(environment_name)


def set_current_environment(new_current_environment: str):
    myProject = Project.instance()

    try:
        myProject.set_current_environment(new_current_environment)
    except Exception as e:
        print(f'Could not set {new_current_environment} as the environment')
        return

    print(f"Set Current Environment -> {new_current_environment}")


def create_environment(new_environment_name: str):
    myProject = Project.instance()

    myProject.create_environment(new_environment_name)

    print(f"Created Environment -> {new_environment_name}")


def settings_information(key: str = None, new_value: str = None, all: bool = False):
    myProject = Project.instance()
    

    if all and not new_value:
        raise Exception('Must use --all with --new-value')


    if not new_value:
        settings_info = myProject.get_settings_info()
        settings_dict = settings_info.dict()
        print(f'Settings info for Environment {myProject.get_current_environment_name()}:')
        if key:
            #Print desired key
            if key not in settings_dict:
                raise Exception(f"Key {key} not in settings information {settings_info}")

            print(f"    {key} -> {settings_dict.get(key)}")

        else:
            # Print all values
            for key, value in settings_dict.items():
                print(f"    {key} -> {value}")

    
    else:
        if not key:
            raise Exception('Must use --new-value with --key')

        if not all:
            settings_info = myProject.get_settings_info()

            Confirm.ask(f"Are you sure you want to update {key} to {new_value} for the current environment ({myProject.get_current_environment_name()})?")

            setattr(settings_info, key, new_value)

            myProject.update_settings_info(settings_info)
            print(f"Updated {key} -> {new_value}")


        else:
            Confirm.ask(f"Are you sure you want to update {key} to {new_value} for all environments?")
            for environment_name in myProject.get_all_environment_names():
                settings_info = myProject.get_settings_info(environment_name)

                setattr(settings_info, key, new_value)

                myProject.update_settings_info(settings_info, environment_name)
                print(f"Updated ({environment_name}) {key} -> {new_value}")

    




import ast
import os
import symtable
from enum import Enum
import tokenize
import re
import hashlib
from typing import List

from sortedcontainers import SortedList

INCLUDE_REGEX = '^#include <(\w+)>$'

def _hash_line_numbers(line_no_1: int, line_no_2: int) -> str:
    s_numbers = f"{line_no_1}:{line_no_2}"
    return int.from_bytes( hashlib.md5(s_numbers.encode()).digest(), byteorder='big')


class parsed_function():
    """
        This class represents the information for a parsed function. It mostly is used to house a function that can keep track of the line numbers.
    """

    def __init__(self, name):
        self.name: str = name
        self.needed_line_numbers = SortedList(key=lambda x: x[0])

        # set of imported packages
        self.imported_packages: str = set()

        # some packages come with the lambda runtime so they do not need to be packaged into a layer so we need to make a new set with 
        # just packages we need to actually package into layers
        # EX: 'os' does not need to be packaged into a layer
        self.needed_imports = set()

        self.needed_imports_hash = ""

    def __eq__(self):
        return self.name == self.name

    def add_line_numbers(self, line_no):
        # Use a sorted list
        self.needed_line_numbers.add(line_no)

    def get_line_numbers(self):
        return self.needed_line_numbers

    def get_line_numbers_serializeable(self):
        return list(self.needed_line_numbers)

    def add_import(self, global_import_obj):
        #self.add_line_numbers(global_import_obj.get_line_no())
        # original package will be how it is denoted in the import statement
        # for absolute packages we only want the top level name (absolutes don't start with '.')
        # for relative packages we want the entire name (relatives start with '.')
        top_level_package_name = global_import_obj.original_package.split('.')[0] if not global_import_obj.original_package.startswith('.') else global_import_obj.original_package
        self.imported_packages.add(top_level_package_name)


class GlobalStatementType(Enum):
    STANDARD = 1
    FUNCTION = 2
    IMPORT = 3
    CLASS_DEF = 4


class GlobalStatement():
    """
        This class represents the extra information for the global statements of a file. A global statement are the first children in the 
        ast for a file. 
    """
    def __init__(self, node, line_no, symbols):
        self.line_no = line_no

        self.set_symbols(symbols)

        self.node = node

        self.hashed = _hash_line_numbers(self.line_no[0], self.line_no[1])

        # The type of statement
        self.statement_type = GlobalStatementType.STANDARD


    def __hash__(self):
        return self.hashed

    def __eq__(self, other):
        return self.hashed == other.hashed

    def __repr__(self):
        return f"<statement at {self.line_no[0]} thru {self.line_no[1]}>"

    def set_symbols(self, symbols: List[str]):
        self.symbols = symbols

    def get_symbols(self) -> List[str]:
        return self.symbols

    def get_line_no(self):
        return self.line_no

    def get_type(self):
        return self.statement_type


class ImportStatement(GlobalStatement):
    # The original package that the package is from
    # ex: import pandas as pd... original_package = 'pandas'
    original_package = ""

    # The actual symbol the package is import as
    # ex: import pandas as pd... as_symbol = 'pd'
    as_symbol = ""

    # Is local file or pkg
    is_local = False

    def __init__(self, node, line_no, symbol_table, asname, pkgname):
        super().__init__(node, line_no, symbol_table)
        self.as_symbol = asname
        self.original_package = pkgname

    def get_type(self):
        return GlobalStatementType.IMPORT


class FunctionStatement(GlobalStatement):

    # function name is a func
    func_name = ""

    has_decremented = False

    def __init__(self, node, line_no, symbol_table, name):
        super().__init__(node, line_no, symbol_table)
        self.func_name = name
        self.has_decremented = False

    def get_type(self):
        return GlobalStatementType.FUNCTION

    def get_function_name(self):
        return self.func_name

    def decrement_line_number(self):
        if self.has_decremented:
            return
        
        
        old_lineno = self.line_no
        self.line_no = [old_lineno[0]+1, old_lineno[1]]
        self.has_decremented = True


class ClassDefinitionStatement(GlobalStatement):

    def __init__(self, node, line_no, symbol_table, name):
        super().__init__(node, line_no, symbol_table)
        self.class_name = name

    def get_type(self):
        return GlobalStatementType.CLASS_DEF

    def get_class_name(self):
        return self.class_name
    



class file_information():
    """
        This class represents a file that needs to have functions parsed out of it. 
    """

    # init method or constructor
    def __init__(self, location):
        if not os.path.isfile(location):
            raise FileNotFoundError(
                f"parser_utils: could not find file at -> {location}")
        
        # Path to the file
        self.file_location = location

        # Source code
        self.src_code = []

        # dict<str, global_statement>: imported symbol name to global statement
        self.imported_symbol_to_global_statement = {}

        # Set of the symbol names
        self.imported_symbols = set()

        # List of parsed function objects
        self.parsed_functions: List[parsed_function] = []

        # dict<str, global_statement>: function name to its global statement
        self.global_functions = {}

        
        # dict<statement, list(str)>: function from statement to symbols
        self.statement_to_symbol = {}

        # dict<str, list(statement)>: dictionary from string name of symbol to global statement
        self.symbol_to_statement = {}

        # list[global_statements]
        self.top_level_statements = []

        # dict<str, int>: dict from label for manual override to the line number of the comment
        self.include_overrides_lineno = {}

        # dict<str, globalobj>: dict from label for manual override to the line number of the comment
        self.include_overrides_glob = {}

        # symbol table for the file
        self.symbol_table = None

        
        # abstract syntax tree for the file
        self.ast = ""

        with open(location, 'r') as fh:
            self.src_code = fh.readlines()
            fh.seek(0)

            for toktype, tok, start, end, line in tokenize.generate_tokens(fh.readline):
                # we can also use token.tok_name[toktype] instead of 'COMMENT'
                # from the token module 
                if toktype == tokenize.COMMENT:
                    self._evaluate_comment(line, start)

        try:
            self.symbol_table = symtable.symtable("".join(self.src_code),
                                                  location, 'exec')
        except SyntaxError as e:
            print(e)

        try:
            self.ast = ast.parse("".join(self.src_code))
        except Exception as e:
            print(e)

    def _evaluate_comment(self, line, start):
        #print(f"COMMENT {start} {line}")

        match_obj = re.match(INCLUDE_REGEX, line)

        if match_obj:
            if len(match_obj.groups()) == 1:
                self.include_overrides_lineno[match_obj.groups()[0]] = start[0]


    def get_source_code(self):
        return "".join(self.src_code)

    def get_lines_of_source_code(self, start_line, end_line):
        return "".join(self.src_code[(start_line - 1):(end_line)])

    def get_symbol_table(self):
        return self.symbol_table

    def get_ast(self):
        return self.ast

    def add_global_statement(self, global_statement):
        self.top_level_statements.append(global_statement)
        self.statement_to_symbol[global_statement] = set()

    def get_global_statements(self) -> List[GlobalStatement]:
        return self.top_level_statements

    def get_file_length(self):
        return len(self.src_code)

    def add_global_function(self, name, global_obj):
        self.global_functions[name] = global_obj
        self.add_global_statement(global_obj)

    def add_global_import(self, global_obj):
        self.imported_symbol_to_global_statement[
            global_obj.as_symbol] = global_obj
            
        self.add_global_statement(global_obj)

    def add_class_definition(self, name, global_obj):
        self.global_functions[name] = global_obj
        self.add_global_statement(global_obj)

    def add_parsed_functions(self, func):
        self.parsed_functions.append(func)

    def set_imported_symbols(self, symbols):
        self.imported_symbols = symbols



from core.constructs.commands import BaseCommandContainer


class SimpleCommandContainer(BaseCommandContainer):
    help = """
These are the commands that are available for use on the simple xlambda.

execute: Trigger a deployed function with a given event and context
logs: Get the logs from a deployed function


"""



"""Structure that deploys the resources onto the cloud 

"""

from typing import Dict, Union, List, Optional, Set, Callable

from .resource import Resource_Difference

from core.constructs.output_manager import OutputTask


class CloudMapper:
    """
    A Cloud Mapper is the construct responsible for directly interacting with the Cloud Provider and managing resource state.
    """

    def __init__(self, resource_to_handler: Dict[str, Callable]) -> None:
        self.resource_to_handler = resource_to_handler

    def get_namespaces(self) -> List[str]:
        raise NotImplementedError

    def deploy_resource(self, transaction_token: str, namespace_token: str, resource_diff: Resource_Difference, previous_output: Dict, output_task: OutputTask) -> Dict:
        raise NotImplementedError

    def get_available_resources(self) -> Set[str]:
        raise NotImplementedError

    def get_resource_to_handler(self) -> Dict[str, Callable]:
        raise NotImplementedError

